import sys
import logging
import os
import argparse
import json

logger = None



def __insert_sys_paths():
    # Insert path to folder containing script_tools
    here = os.path.abspath(os.path.dirname(__file__))
    scripts_path = os.path.abspath(os.path.join(here, "..", ".."))
    if scripts_path not in sys.path:
        sys.path.insert(0, scripts_path)
    
    from script_tools import insert_holado_helper  # @UnresolvedImport
    insert_holado_helper()
    
    
def __main(args, cwd):
    from holado_python.standard_library.csv import CsvManager
    from holado_core.common.exceptions.technical_exception import TechnicalException
    from holado_core.common.tables.converters.table_converter import TableConverter
    
    input_filepath = os.path.join(cwd, args.input_filepath)
    
    input_table = CsvManager.table_with_content_of_CSV_file(input_filepath, with_header=False, encoding='utf-8-sig', delimiter=';')
    summary_ranking_by_category = __extract_summary_ranking_by_category(input_table)
    detailed_ranking_by_category = __extract_detailed_ranking_by_category(input_table)
    
    if args.output_filepath is None:
        print("Summary ranking:")
        ranking_by_category_json = {k:TableConverter.convert_table_with_header_to_dict_list(summary_ranking_by_category[k]) for k in summary_ranking_by_category}
        print(json.dumps(ranking_by_category_json, indent=4))
        
        print("\n\nDetailed ranking:")
        ranking_by_category_json = {k:TableConverter.convert_table_with_header_to_dict_list(detailed_ranking_by_category[k]) for k in detailed_ranking_by_category}
        print(json.dumps(ranking_by_category_json, indent=4))
    else:
        output_filepath = os.path.join(cwd, args.output_filepath)
        _, ext = os.path.splitext(output_filepath)
        if ext.lower() == '.html':
            __create_html_ranking(output_filepath, summary_ranking_by_category, detailed_ranking_by_category)
        else:
            raise TechnicalException(f"Unmanaged output with extension '{ext}'")
    
def __extract_summary_ranking_by_category(table):
    res = {}
    
    categories_indexes = [(5, 0), (5, 4), (5, 8), (5, 12), (5, 16), (5, 20), (5, 24)]
    for ci in categories_indexes:
        category, ranking_table = __extract_category_ranking_table(table, *ci)
        res[category] = ranking_table
    
    return res
    
def __extract_detailed_ranking_by_category(table):
    res = {}
    
    categories_indexes = [(45, 0), (45, 8), (45, 16), (45, 24), (45, 32), (45, 40), (45, 48)]
    for ci in categories_indexes:
        category, ranking_table = __extract_category_ranking_table(table, *ci)
        res[category] = ranking_table
    
    return res

def __extract_category_ranking_table(table, row_ref_index, col_ref_index):
    from holado_core.common.tables.converters.table_converter import TableConverter
    
    category = table.rows[row_ref_index].cells[col_ref_index].content
    
    row_first_index = row_ref_index + 1
    col_first_index = col_ref_index
    
    row_last_index = row_first_index
    while row_last_index + 1 < table.nb_rows:
        if table.rows[row_last_index+1].cells[col_first_index+1].content in ['', '0']:
            break
        else:
            row_last_index += 1
    
    col_last_index = col_first_index
    while col_last_index + 1 < table.nb_columns:
        if table.rows[row_first_index].cells[col_last_index+1].content in ['']:
            break
        else:
            col_last_index += 1
    
    res_table = TableConverter.extract_table_with_header_from_table(table, row_first_index, row_last_index, col_first_index, col_last_index)
    for row in res_table.rows:
        row.cells[0].content = int(row.cells[0].content)
    res_table.sort(names=["Rang"])
    
    return category, res_table





def __create_html_ranking(output_filepath, summary_ranking_by_category, detailed_ranking_by_category):
    with open(output_filepath, 'w') as fout:
        __write_html_begin(fout)
        __write_html_ranking(fout, "Classement général", summary_ranking_by_category)
        __write_html_ranking(fout, "Détails par manche", detailed_ranking_by_category)
        __write_html_end(fout)
    
def __write_html_begin(fout):
    __writeline(fout, "<html>")
    __writeline(fout, "<body>")
    
def __write_html_ranking(fout, title, ranking_by_category):
    __writeline(fout, f"<h1>{title}</h1>")
    for category in ranking_by_category:
        __write_html_ranking_category(fout, category, ranking_by_category[category])
    
def __write_html_ranking_category(fout, category, ranking_table):
    __writeline(fout, f"<h2>{category}</h2>")
    
    __writeline(fout, f"<table>")
    
    __writeline(fout, f"<tr>")
    for hc in ranking_table.header:
        __writeline(fout, f"<th>{hc.content}</th>")
    __writeline(fout, f"</tr>")
    
    for row in ranking_table.rows:
        __writeline(fout, f"<tr>")
        for c in row:
            __writeline(fout, f"<td>{c.content}</td>")
        __writeline(fout, f"</tr>")
        
    __writeline(fout, f"</table>")
    
def __write_html_end(fout):
    __writeline(fout, "</body>")
    __writeline(fout, "</html>")
    
def __writeline(fout, line):
    fout.write(line + "\n")
    
def _parse_arguments():
    descr = """Roadmap
    
    Usage:
        1. Generate a template file for the desired action with parameter '--output-template' as in example 1.
            It is possible to specify the name of the output file with parameter '-o'.
        2. Copy the template line in order to have one line per action to do, and modify the template file with desired values. 
        3. Process actions described in a CSV file with parameter '-i'.
            If an error occurs at a line, execution is stopped.
        4. See result in output file. If parameter '-o' is not specified, the output filename is generated from parameter '-i'.
    
    In template files, a default value is set in each column. 
    For columns with enum values, the default value is a combination of each possible enum values separated by a '|', but only one value is possible.
    
    
    In input and output files, the first column 'STATUS' describes the status of execution of each line:
        - 'TODO' (or ''): the line is to process (in output file, it means that execution was stopped on an error with a previous line)
        - 'SUCCESS': the line was in success
        - 'ERROR': the line was in error
        - others: the line is not processed, you can specify your own status to skip this line
    In template files, this column is automatically initialized with status 'TODO'.
    Any output file can be used as input file, the lines that are in status 'ERROR' or 'TODO' will be processed again.
    
    Example 1: Output CSV template
            
            roadmap.sh --output-template -o {action}.csv
            
    Example 2: Process with given input CSV
            
            roadmap.sh -i {action}.csv
            
    """
    
    parser = argparse.ArgumentParser(description=descr, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-i', '--input-filepath', dest='input_filepath',
                   help='Input file path')
    
    parser.add_argument('-o', '--output-filepath', dest='output_filepath',
                   help='Output file path')
    
    parser.add_argument('-ll', '--log-level', choices=['ERROR', 'WARNING', 'INFO', 'DEBUG', 'TRACE'], dest='log_level', default='WARNING',
                   help='Log level')
    
    parser.add_argument('-lf', '--log-in-file', dest='log_in_file', default=False, action="store_true",
                   help='If specified, log in file rather than console')
    
    res = parser.parse_args()
    
    return res


if __name__ == "__main__":
    __insert_sys_paths()
    from holado_helper.script.initialize_script import initialize  # @UnresolvedImport
    
    script_path = os.path.abspath(os.path.dirname(__file__))
    
    args = _parse_arguments()
    
    # Initialize HolAdo framework
    cwd = initialize(args.log_level, script_path, log_in_file=args.log_in_file)
    
    from holado_core.common.exceptions.functional_exception import FunctionalException
    
    try:
        __main(args, cwd)
    except FunctionalException as exc:
        logging.error(f"Action is interrupted on error: {exc}")
        sys.exit(1)
    except SystemExit as exc:
        raise exc
    except:
        logging.exception("An error occured")
        sys.exit(1)
    
    sys.exit(0)
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado.common.handlers.object import DeleteableObject, Object

logger = None

def initialize_logger():
    global logger
    logger = logging.getLogger(__name__)


class Context(DeleteableObject):
    """
    @summary: Mother class for any context class
    """
    
    def __init__(self, name, with_post_process=False, persisted_kwargs=None):
        super().__init__(name)
        self.on_delete_call_gc_collect = True
        
        self.__objects = {}
        self.__on_delete_objects = []
        
        # Post process management
        if with_post_process:
            self.__persisted_kwargs = persisted_kwargs if persisted_kwargs is not None else {}
            self.__persisted_method_to_call_manager_inst = None
            self.__post_process_funcs = []
        
    @property
    def __with_post_process(self):
        return hasattr(self, '_Context__persisted_method_to_call_manager_inst')
    
    def __verify_is_with_post_process(self):
        if not self.__with_post_process:
            from holado_core.common.exceptions.technical_exception import TechnicalException
            raise TechnicalException(f"[{self.name}] This context doesn't manage post process")
        
    @property
    def __persisted_method_to_call_manager(self):
        self.__verify_is_with_post_process()
        
        if self.__persisted_method_to_call_manager_inst is None:
            from holado.common.context.session_context import SessionContext
            from holado_core.common.resource.persisted_method_to_call_manager import PersistedMethodToCallManager
            self.__persisted_method_to_call_manager_inst = PersistedMethodToCallManager(scope_name=self.name, **self.__persisted_kwargs)
            self.__persisted_method_to_call_manager_inst.initialize(SessionContext.instance().resource_manager, SessionContext.instance().expression_evaluator)
            self.__persisted_method_to_call_manager_inst.ensure_persistent_db_exists()
        return self.__persisted_method_to_call_manager_inst
        
    def get_context_name(self):
        return Object.name(self)
    
    def _delete_object(self):
        # Remove all objects
        self.remove_all_objects()
    
    def remove_all_objects(self):
        from holado_core.common.exceptions.functional_exception import FunctionalException
        from holado_core.common.exceptions.technical_exception import TechnicalException
        from holado_core.common.tools.tools import Tools
        from holado_python.standard_library.typing import Typing
        
        keys = set(self.__objects.keys())
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Removing {len(keys)} objects: {keys}")
        for index, key in enumerate(keys):
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Removing object {index+1}/{len(keys)}: '{key}' (type: {Typing.get_object_class_fullname(self.get_object(key))})")
            try:
                self.remove_object(key)
            except FunctionalException as exc:
                raise TechnicalException() from exc
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Finished to remove {len(keys)} objects: {keys}")
                
    def remove_object(self, name):
        if not self.has_object(name):
            from holado_core.common.exceptions.technical_exception import TechnicalException
            raise TechnicalException("Context doesn't contain object '{}'".format(name))
        self._remove(name, True)
        
    def _remove(self, name, remove_from_lifetime_manager):
        from holado_core.common.tools.tools import Tools
        from holado_python.standard_library.typing import Typing
        
        obj = self.get_object(name)
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Removing object '{name}' (type: {Typing.get_object_class_fullname(obj)})")
        if name in self.__on_delete_objects:
            if isinstance(obj, DeleteableObject):
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"[{self.name}] Deleting object '{name}'")
                try:
                    obj.delete_object()
                except Exception:
                    logger.exception(f"[{self.name}] Catched exception while deleting object '{name}'")
                else:
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"[{self.name}] Deleted object '{name}'")
            self.__on_delete_objects.remove(name)
            
        if self.has_object(name):
            del self.__objects[name]
#         if remove_from_lifetime_manager:
#             from holado.common.context.session_context import SessionContext
#             SessionContext.instance().get_object_lifetime_manager().remove_object_lifetime(name, self)
        
    def set_object(self, name, obj, raise_if_already_set = True):
        from holado_core.common.exceptions.functional_exception import FunctionalException
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        if obj is None:
            if self.has_object(name):
                try:
                    self.remove_object(name)
                except FunctionalException as e:
                    raise TechnicalException() from e
        else:
            if raise_if_already_set and self.has_object(name):
                raise TechnicalException(f"[{self.name}] Context already contains object '{name}'")
            self._put(name, obj)
                    
    def _put(self, name, obj):
        self.__objects[name] = obj
        self.__on_delete_objects.append(name)
        
    def get_object(self, name):
        if not self.has_object(name):
            from holado_core.common.exceptions.technical_exception import TechnicalException
            raise TechnicalException(f"[{self.name}] Context doesn't contain object '{name}'")
        return self.__objects[name]

    def has_object(self, name):
        return name in self.__objects
    
    def get_object_name(self, obj):
        for name, value in self.__objects.items():
            if value is obj:
                return name
        return None
        
    def remove_lifetimed_object(self, object_id):
        self._remove(object_id, False)
        
    def get_or_create_initilizable_object_by_class(self, object_class):
        raise NotImplementedError()
        
    def get_or_create_object_by_class(self, object_class):
        raise NotImplementedError()
    
    def get_object_getter_eval_string(self, obj, raise_not_found=True):
        from holado_python.standard_library.typing import Typing
        
        name = self.get_object_name(obj)
        if name is not None:
            return f"{Typing.get_object_class_fullname(self)}.instance().get_object('{name}')"
        
        if raise_not_found:
            from holado_core.common.exceptions.element_exception import ElementNotFoundException
            raise ElementNotFoundException(f"[{self.name}] Failed to find object of id {id(obj)}")
        else:
            return None

    
    
    ### Post process management
    
    def add_post_process(self, func):
        self.__verify_is_with_post_process()
        
        success = self.__persist_function(func)
        if not success:
            logger.debug(f"[{self.name}] Add post process in memory, persistence in DB has failed (post process function: {func})")
            self.__post_process_funcs.append(func)
    
    def __persist_function(self, func):
        from holado.common.context.session_context import SessionContext
        from holado_core.common.exceptions.technical_exception import TechnicalException
        from holado_core.common.tools.tools import Tools
        from holado_python.standard_library.typing import Typing
        
        try:
            if Typing.is_function(func._target):
                func_qualname = Typing.get_function_fullname(func._target)
                self.__persisted_method_to_call_manager.add_function_to_call(
                    func_qualname, args_list=func._args, kwargs_dict=func._kwargs, 
                    use="post_process", use_index=None)
                return True
            elif Typing.is_method(func._target):
                meth_func = Typing.get_method_function(func._target)
                func_qualname = Typing.get_function_fullname(meth_func)
                meth_obj = Typing.get_method_object_instance(func._target)
                self_getter_eval_str = SessionContext.instance().get_object_getter_eval_string(meth_obj, raise_not_found=False)
                if self_getter_eval_str is not None:
                    self.__persisted_method_to_call_manager.add_method_to_call(
                        func_qualname, self_getter_eval_str, args_list=func._args, kwargs_dict=func._kwargs, 
                        use="post_process", use_index=None)
                    return True
            else:
                raise TechnicalException(f"[{self.name}] Unmanaged target type '{Typing.get_object_class_fullname(func._target)}'")
        except TechnicalException as exc:
            raise exc
        except Exception as exc:
            logger.warning(f"[{self.name}] Failed to persist function {func}: {Tools.represent_exception(exc)}")
        return False
        
    def do_post_processes(self):
        self.__verify_is_with_post_process()
        
        # First call functions that were not persisted
        for func in self.__post_process_funcs:
            try:
                func.run()
            except Exception as exc:
                # logger.error(f"Error while post processing [{func}]: {exc}")
                logger.exception(f"[{self.name}] Error while post processing [{func}]: {exc}")
        
        # Call persisted functions and methods
        try:
            self.do_persisted_post_processes()
        except Exception as exc:
            logger.exception(f"[{self.name}] Error while scenario post processing persisted methods: {exc}")
        
    def do_persisted_post_processes(self):
        """Call functions and methods persisted by a previous context instance
        Note: This is useful especially when a context was deleted (like on process kill) before post processes could be performed
        """
        self.__verify_is_with_post_process()
        
        self.__persisted_method_to_call_manager.call_functions_and_methods(use="post_process")
        
    def get_number_of_persisted_post_processes_to_perform(self):
        return self.__persisted_method_to_call_manager.get_number_of_functions_and_method_to_call()






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from builtins import object
import threading
import gc
from holado.common.handlers.undefined import default_value

logger = None
default_collect_periodicity = 10

def initialize_logger():
    global logger
    logger = logging.getLogger(__name__)


class GcManager(object):
    """
    Manage Garbage Collector
        
    Note: When working with HolAdo, it is highly recommended to call collect_periodically method at process start in order to avoid deadlocks.
          This is done by default in holado.initialize method.
    Explanation:
        HolAdo is using object finalizers (define __del__ methods) to properly release objects resources.
        But in current garbage collector default mecanism, objects are technically finalized (call of __del__) in any yielded thread.
        Thus, depending on what thread is doing when an object is finalized, a deadlock can occur.
        Example: current thread is logging a message, garbage collector is finalizing an object in this thread when logger flushes its stream,
                 and object finalizer tries to log an other message => it results to a deadlock in logging library in its lock mecanism.
    """
    
    # Manage running and pended collect
    __collect_run_lock = threading.Lock()
    __is_collect_running = False
    __is_collect_pending = False
    __duration_until_pended_collect = default_collect_periodicity
    
    # Manage scheduled collect as timer
    __collect_timer_lock = threading.RLock()
    __collect_periodicity = None
    __collect_timer = None
    __collect_timer_at = None
    
    @classmethod
    def is_collect_threaded(cls):
        return cls.__collect_periodicity is not None
    
    @classmethod
    def get_collect_periodicity(cls):
        return cls.__collect_periodicity
    
    @classmethod
    def collect_periodically(cls, collect_periodicity=default_value):
        """
        Call this method to periodically process gargage collector collect.
        By calling this method, default behaviour of garbage collector is disabled.
        """
        gc.disable()
        with cls.__collect_timer_lock:
            cls.__collect_periodicity = default_collect_periodicity if collect_periodicity is default_value else collect_periodicity
        cls.__schedule_next_timer(cls.__collect_periodicity)
    
    @classmethod
    def collect(cls, max_duration_until_next_collect_in_seconds=default_value):
        max_duration = default_collect_periodicity if max_duration_until_next_collect_in_seconds is default_value else max_duration_until_next_collect_in_seconds
        
        with cls.__collect_run_lock:
            # If a collect is already pended, update duration until pended collect
            if cls.__is_collect_pending:
                if max_duration < cls.__duration_until_pended_collect:
                    cls.__duration_until_pended_collect = max_duration
                return
            # If a collect is already running, pend a collect
            if cls.__is_collect_running:
                cls.__is_collect_pending = True
                cls.__duration_until_pended_collect = max_duration
                return
        
        # Else schedule next collect
        with cls.__collect_timer_lock:
            if cls.__collect_timer is not None:
                cls.__schedule_next_timer(max_duration_until_next_collect_in_seconds)
            else:
                cls.__schedule_next_timer(0)
        
    @classmethod
    def __collect(cls):
        # Start running
        with cls.__collect_run_lock:
            cls.__is_collect_running = True
        
        # Process collect
        gc.collect()
        
        # Schedule next timer
        duration_until_next_collect = None
        with cls.__collect_run_lock:
            if cls.__is_collect_pending:
                duration_until_next_collect = cls.__duration_until_pended_collect
        with cls.__collect_timer_lock:
            if cls.__collect_periodicity is not None:
                if duration_until_next_collect is None:
                    duration_until_next_collect = cls.__collect_periodicity
                else:
                    duration_until_next_collect = min(duration_until_next_collect, cls.__collect_periodicity)
            
            cls.__schedule_next_timer(duration_until_next_collect)
        
        # Stop running
        with cls.__collect_run_lock:
            cls.__is_collect_running = False
            # Reinitialize pending variables
            cls.__is_collect_pending = False
            cls.__duration_until_pended_collect = default_collect_periodicity
            
    @classmethod
    def __schedule_next_timer(cls, duration_until_next_collect_seconds):
        import holado_multitask.multithreading.timer as mmt
        from holado_multitask.multitasking.multitask_manager import MultitaskManager
        from holado_python.common.tools.datetime import DateTime, DurationUnit
        from holado_python.common.enums import ArithmeticOperator
        
        if duration_until_next_collect_seconds is default_value:
            duration_until_next_collect_seconds = default_collect_periodicity
        
        with cls.__collect_timer_lock:
            if duration_until_next_collect_seconds is not None:
                now = DateTime.now()
                next_timer_at = DateTime.apply_delta_on_datetime(now, ArithmeticOperator.Addition, duration_until_next_collect_seconds, DurationUnit.Second)
                
                if cls.__collect_timer_at is not None and cls.__collect_timer_at > now and next_timer_at > cls.__collect_timer_at:
                    # Current scheduled collect is not running and will occur before newly schedule, it is preserved
                    return
                elif cls.__collect_timer_at is not None:
                    # Current scheduled collect is canceled and replaced
                    cls.__collect_timer.cancel()
                
                # Schedule timer with given duration
                cls.__collect_timer = mmt.Timer("GcManager-collect", duration_until_next_collect_seconds, cls.__collect, register_scope=MultitaskManager.SESSION_SCOPE_NAME)
                cls.__collect_timer.start()
                cls.__collect_timer_at = next_timer_at
            else:
                cls.__collect_timer = None
                cls.__collect_timer_at = None




# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


import logging
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_core.common.block.function import Function
from holado_test.common.context.scenario_context import ScenarioContext
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado_test.scenario.step_tools import StepTools
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


def __get_scenario_context() -> ScenarioContext:
    return SessionContext.instance().get_scenario_context()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()



@Given(r"at end of scenario, call steps(?: \(format with: (?P<format_with>{Variable}(?:, {Variable})*)\))?")
def step_impl(context, format_with):
    format_with = StepTools.evaluate_variable_name(format_with)
    steps = BehaveStepTools.get_step_multiline_text(context, eval_params=EvaluateParameters.nothing())
    
    if format_with is not None:
        steps = StepTools.format_steps_with(steps, format_with.split(', '))
    
    func = Function(execute_steps, steps)
    __get_scenario_context().add_post_process(func)


@Given(r"(?P<var_name>{Variable}) = last step duration")
def step_impl(context, var_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    
    step_context = __get_scenario_context().get_step(-2)
    res = step_context.duration
    if res is None:
        raise TechnicalException(f"Failed to get last step duration: {step_context}")
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = last step start datetime")
def step_impl(context, var_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    
    step_context = __get_scenario_context().get_step(-2)
    res = step_context.start_datetime
    if res is None:
        raise TechnicalException(f"Failed to get last step start datetime: {step_context}")
    
    __get_variable_manager().register_variable(var_name, res)
    
    __get_variable_manager().register_variable(var_name, step_context.start_datetime)

@Given(r"(?P<var_name>{Variable}) = last step end datetime")
def step_impl(context, var_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    
    step_context = __get_scenario_context().get_step(-2)
    res = step_context.end_datetime
    if res is None:
        raise TechnicalException(f"Failed to get last step end datetime: {step_context}")
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = defined user data (?P<user_data_name>{Str})(?: \(default: (?P<default_value>{Any})\))?")
def step_impl(context, var_name, user_data_name, default_value):
    var_name = StepTools.evaluate_variable_name(var_name)
    user_data_name = StepTools.evaluate_scenario_parameter(user_data_name)
    default_value = StepTools.evaluate_scenario_parameter(default_value)
    
    if user_data_name in context.config.userdata:
        res = context.config.userdata[user_data_name]
    else:
        res = default_value
    
    __get_variable_manager().register_variable(var_name, res)
    
    






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_ui.ui.inspectors.ui_inspector import UIInspector
import weakref
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_multitask.multithreading.reflection.sys import get_current_function_name
from holado_ui.ui.gui.windows.gui_window import GUIWindow
from holado_ui.ui.gui.handlers.zone_gui_context import ZoneGUIContext,\
    InformationTypes
import abc
from holado_core.common.inspectors.tools.inspect_parameters import InspectParameters
from holado_ui.ui.inspectors.tools.ui_inspect_context import UIInspectContext
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)



class GUIInspector(UIInspector):
    """ Base class for GUI inspector.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, module_name):
        super().__init__(module_name)
        
        self.__window_weakref = weakref.ref(None)
    
    def initialize(self, window:GUIWindow):
        if window is None:
            raise TechnicalException("Window is mandatory")
        super().initialize(window.driver)
        
        self.__window_weakref = weakref.ref(window)
        
    @property
    def window(self):
        res = self.__window_weakref()
        if res is None:
            raise TechnicalException("Window has been cleared")
        return res
        
    @UIInspector.inspect_builder.getter  # @UndefinedVariable
    def inspect_builder(self):
        """
        Prior inspect builder in window than in inspector
        @return Inspect builder instance
        """
        # Priors window one if existing 
        return self.get_inspect_builder(False)

    def get_inspect_builder(self, from_window):
        """
        Used internally to prior inspect builder in window than in inspector
        @param from_window If request comes from window
        @return Inspect builder instance
        """
        res = None
        
        if not from_window:
            res = self.window.get_inspect_builder(True)
        
        if res is None:
            res = super().inspect_builder
        
        return res

    @UIInspector.find_builder.getter  # @UndefinedVariable
    def find_builder(self):
        """
        Prior find builder in window than in inspector
        @return Find builder instance
        """
        # Priors window one if existing 
        return self.get_find_builder(False)

    def get_find_builder(self, from_window):
        """
        Used internally to prior find builder in window than in inspector
        @param from_window If request comes from window
        @return Find builder instance
        """
        res = None
        
        if not from_window:
            res = self.window.get_find_builder(True)
        
        if res is None:
            res = super().find_builder
        
        return res
    
    def _get_finder_ui_context_element(self, ui_context, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        if ui_context is None:
            return None
        
        if isinstance(ui_context, ZoneGUIContext):
            if ui_context.information_type == InformationTypes.Label:
                return self.get_finder_zone(ui_context.information, inspect_context, inspect_parameters)
            else:
                raise TechnicalException(f"Unmanaged zone information: type='{ui_context.information_type.name}'  information='{ui_context.information}'")
        else:
            raise TechnicalException(f"Unmanaged UI context type '{Typing.get_object_class_fullname(ui_context)}'")
    
    def get_finder_label(self, label, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param label Text
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for an element with given label text.
        """
        return self._get_finder_from_modules(get_current_function_name(), [label], None, inspect_context, inspect_parameters)

    def get_finder_popup(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder for a popup
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        return self._get_finder_from_modules(get_current_function_name(), None, None, inspect_context, inspect_parameters)
    
    def get_finder_popup_title(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder for a popup title
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        return self._get_finder_from_modules(get_current_function_name(), None, None, inspect_context, inspect_parameters)
    
    def get_finder_symbol(self, expected_symbol, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param expected_symbol Symbol as string
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for element corresponding to given symbol
        """
        return self._get_finder_from_modules(get_current_function_name(), [expected_symbol], None, inspect_context, inspect_parameters)
    
    def get_finder_table(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for a table element
        """
        return self._get_finder_from_modules(get_current_function_name(), None, None, inspect_context, inspect_parameters)
    
    def get_finder_table_row(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for a table row
        """
        return self._get_finder_from_modules(get_current_function_name(), None, None, inspect_context, inspect_parameters)
    
    def get_finder_table_cell(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for a table cell
        """
        return self._get_finder_from_modules(get_current_function_name(), None, None, inspect_context, inspect_parameters)
    
    def get_finder_text(self, text, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param text Text to find
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder of element with given text
        """
        return self._get_finder_from_modules(get_current_function_name(), [text], None, inspect_context, inspect_parameters)
    
    def get_finder_text_containing(self, text, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param text Text
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for element with content containing given text
        """
        return self._get_finder_from_modules(get_current_function_name(), [text], None, inspect_context, inspect_parameters)
    
    def get_finder_text_element(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder of element displaying text
        """
        return self._get_finder_from_modules(get_current_function_name(), None, None, inspect_context, inspect_parameters)
    
    def get_finder_zone(self, zone_name, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder of zone element for a given zone name.
        @param zone_name Zone name
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        return self._get_finder_from_modules(get_current_function_name(), [zone_name], None, inspect_context, inspect_parameters)
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.tools.comparators.comparator import Comparator,\
    CompareOperator
from holado_core.common.tools.tools import Tools
from holado.common.context.session_context import SessionContext
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class StringComparator(Comparator):

    def __init__(self, text_interpreter=None, text_verifier=None):
        super().__init__("string")
        self.__text_interpreter = text_interpreter
        self.__text_verifier = text_verifier
        
    def __get_text_interpreter(self):
        if self.__text_interpreter is not None:
            return self.__text_interpreter
        elif SessionContext.instance().has_scenario_context():
            return SessionContext.instance().get_scenario_context().get_text_interpreter()
        else:
            return SessionContext.instance().text_interpreter
    
    def __get_text_verifier(self):
        if self.__text_verifier is not None:
            return self.__text_verifier
        elif SessionContext.instance().has_scenario_context():
            return SessionContext.instance().get_scenario_context().get_text_verifier()
        else:
            return SessionContext.instance().text_verifier
    
    def _compare_by_operator(self, obj_1, operator: CompareOperator, obj_2, is_obtained_vs_expected = True, raise_exception = True, redirect_to_equals=True):
        if operator == CompareOperator.Equal:
            return self.__verify_by_text_interpreter(obj_1, obj_2, is_obtained_vs_expected, raise_exception)
        else:
            i_obj_1 = self.__get_text_interpreter().interpret(obj_1)
            i_obj_2 = self.__get_text_interpreter().interpret(obj_2)
            super()._compare_by_operator(i_obj_1, operator, i_obj_2, is_obtained_vs_expected, raise_exception, redirect_to_equals=redirect_to_equals)
    
    def _convert_input(self, obj, name):
        if isinstance(obj, str):
            res = obj
        else:
            res = str(obj)
        return res
    
    def __verify_by_text_interpreter(self, str_1, str_2, is_obtained_vs_expected = True, raise_exception = True):
        """
        Compare by equality.
        Switch is_obtained_vs_expected:
            - True: value 1 is considered obtained ; value 2 is considered expected ; only expected value is interpreted
            - False: value 1 is considered expected ; value 2 is considered obtained ; only expected value is interpreted
            - None: value 1 is compared to value 2 ; both values are interpreted
        """
        if not isinstance(str_1, str):
            if raise_exception:
                raise VerifyException(f"{self._get_name_1(is_obtained_vs_expected).capitalize()} value is not a string: [{str_1}] (type: {Typing.get_object_class_fullname(str_1)})")
            else:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Compare result is false since {self._get_name_1(is_obtained_vs_expected)} value is not a string: [{str_1}] (type: {Typing.get_object_class_fullname(str_1)})")
                return False
        if not isinstance(str_2, str):
            if raise_exception:
                raise VerifyException(f"{self._get_name_2(is_obtained_vs_expected).capitalize()} value is not a string: [{str_2}] (type: {Typing.get_object_class_fullname(str_2)})")
            else:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Compare result is false since {self._get_name_2(is_obtained_vs_expected)} value is not a string: [{str_2}] (type: {Typing.get_object_class_fullname(str_2)})")
                return False
        
        if is_obtained_vs_expected is None:
            # Interpret obtained as TextInterpreter.verify won't do it on obtained parameter
            str_1 = self.__get_text_interpreter().interpret(str_1)
        
        if is_obtained_vs_expected == True or is_obtained_vs_expected is None:
            res = self.__get_text_verifier().verify(str_1, str_2, raise_exception=raise_exception)
        else:
            res = self.__get_text_verifier().verify(str_2, str_1, raise_exception=raise_exception)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("[{}] str_1 equals str_2 => {}\n    str_1 = [{}]\n    str_2 = [{}]".format(Typing.get_object_class(self).__name__, res, str_1, str_2))
        return res
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.context.session_context import SessionContext
from holado_core.common.tools.tools import Tools
import logging
from holado_report.report.reports.base_report import BaseReport
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado_python.common.tools.datetime import DateTime, FORMAT_DATETIME_ISO, FORMAT_TIME_ISO
from holado.common.handlers.undefined import default_value, undefined_argument
from holado.holado_config import Config
# from holado_core.scenario.scenario_duration_manager import ScenarioDurationManager

logger = logging.getLogger(__name__)




class ReportManager(BaseReport):
    """ Manage reports of current session
    """
    TFeatureReport = None
    TStepTools = None
    
    _scenario_status_information_by_uid = {}
    
    def __init__(self):
        super().__init__()
        
        self.__multitask_manager = None
        
        # Auto configuration
        self.configure()
    
    def configure(self):
        from holado_report.report.reports.feature_report import FeatureReport
        ReportManager.TFeatureReport = FeatureReport
        
        from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
        ReportManager.TStepTools = BehaveStepTools
    
    def initialize(self, multitask_manager):
        self.__multitask_manager = multitask_manager
        
    def initialize_reports(self):
        from holado_report.report.builders.detailed_scenario_report_builder import DetailedScenarioReportBuilder
        from holado_report.report.builders.summary_report_builder import SummaryReportBuilder
        from holado_report.report.builders.summary_scenario_report_builder import SummaryScenarioReportBuilder
        from holado_report.report.builders.short_scenario_report_builder import ShortScenarioReportBuilder
        from holado_report.report.builders.failure_report_builder import FailureReportBuilder
        from holado_report.report.builders.summary_scenario_by_category_report_builder import SummaryScenarioByCategoryReportBuilder
        from holado_report.report.builders.summary_by_category_report_builder import SummaryByCategoryReportBuilder
        
        # self.set_execution_historic()
        
        if self.has_report_path:
            # fn = self.get_path("execution_historic.json")
            # self.add_report_builder(ExecutionHistoricReportBuilder(self.execution_historic, fn))
            
            fn = self.get_path("report_summary_scenario_failed.txt")
            self.add_report_builder(SummaryScenarioReportBuilder(fn, exclude_statuses=['Passed']))
            
            fn = self.get_path("report_summary_scenario_failed_by_category.txt")
            self.add_report_builder(SummaryScenarioByCategoryReportBuilder(fn, exclude_categories=['Always Success']))
            
            fn = self.get_path("report_summary_scenario_all.txt")
            self.add_report_builder(SummaryScenarioReportBuilder(fn))
            
            fn = self.get_path("report_failures.xml")
            self.add_report_builder(FailureReportBuilder(fn))
            
            fn = self.get_path("report_short_scenario_failed.txt")
            self.add_report_builder(ShortScenarioReportBuilder(fn, exclude_statuses=['Passed']))
            
            fn = self.get_path(Config.campaign_manager_import_report_name)  # @UndefinedVariable
            self.add_report_builder(DetailedScenarioReportBuilder(fn))
            
            fn = self.get_path("report_summary.txt")
            self.add_report_builder(SummaryReportBuilder(fn))
        
            fn = self.get_path("report_summary_by_category.txt")
            self.add_report_builder(SummaryByCategoryReportBuilder(fn))
        
    @property
    def __feature_reports(self):
        return self.children_reports("feature")
    
    @property
    def current_feature_report(self):
        # if self.__feature_reports:
        #     return self.__feature_reports[-1][1]
        # else:
        #     return None
        if SessionContext.instance().has_feature_context():
            feature_context = SessionContext.instance().get_feature_context()
            if feature_context.has_object("feature_report"):
                return feature_context.get_object("feature_report")
        return None
    
    @property
    def current_scenario_report(self):
        cfr = self.current_feature_report
        if cfr:
            return cfr.current_scenario_report
        else:
            return None
    
    def new_session(self, report_path):
        self.report_path = report_path
        if self.has_report_path:
            SessionContext.instance().path_manager.makedirs(self.report_path, is_directory=True)
        
        self.initialize_reports()
    
    def before_all(self):
        super().before_all()
    
    def before_feature(self, feature_context, feature):
        # Create new feature report
        fr = ReportManager.TFeatureReport(self, feature_context, feature)
        self.add_child_report(fr, "feature", feature.name, feature.filename)
        fr.initialize_reports()
        
        # Process self reports
        super().before_feature(feature_context, feature, feature_report=fr)
        
        # Store feature report in current feature context
        self.__multitask_manager.get_feature_context().set_object("feature_report", fr)
        
    def before_scenario(self, scenario_context, scenario):
        # Create new scenario report in feature report
        sr = self.current_feature_report.new_scenario_report(scenario_context, scenario)
        
        # Process feature report
        self.current_feature_report.before_scenario(scenario_context, scenario, scenario_report=sr)
        
        # Process self reports
        super().before_scenario(scenario_context, scenario, scenario_report=sr)
        
        # Change active log file to scenario one
        self.__enter_current_scenario_log_file()
    
    def before_step(self, step_context, step, step_level):
        super().before_step(step_context, step, step_level)
        
        # Note: A step can be executed without scenario report (ex: post processes after scenario) 
        if self.current_scenario_report:
            self.current_scenario_report.before_step(step_context, step, step_level)
    
    def after_step(self, step_context, step, step_level):
        super().after_step(step_context, step, step_level)
        
        # Note: A step can be executed without scenario report (ex: post processes after scenario) 
        if self.current_scenario_report:
            self.current_scenario_report.after_step(step_context, step, step_level)
    
    def after_scenario(self, scenario):
        self.current_feature_report.after_scenario(scenario, self.current_scenario_report)
        super().after_scenario(scenario, self.current_scenario_report)
        
        # Change active log file to root one
        self.__leave_current_scenario_log_file()
        
        self.current_scenario_report.release_resources()
    
    def after_feature(self, feature):
        super().after_feature(feature, self.current_feature_report)
        
        self.current_feature_report.release_resources()
    
    def after_all(self):
        super().after_all(build_reports=True)
            
        # Create files using execution historic as input
        # fn_eh = self.get_path("execution_historic.json")
        # sdm = ScenarioDurationManager()
        # sdm.import_execution_historic(fn_eh)
        #
        # fn = self.get_path("scenario_durations.csv")
        # scenario_duration_limits = sdm.compute_scenario_duration_limits()
        # sdm.create_file_scenario_duration_limits(fn, scenario_duration_limits)
        #
        # fn = self.get_path("scenario_duration_tags.csv")
        # duration_limit_tags = [(1, "fast"), (5, "rapid"), (60, "slow")]
        # scenario_duration_tags = sdm.compute_scenario_duration_tags(duration_limit_tags, "long", missing_tag=True, new_tag=True, unchanged_tag=True, with_failed=True)
        # sdm.create_file_scenario_duration_tags(fn, scenario_duration_tags)
        
        # Update campaigns stored in test server
        if self._get_test_server_client().is_available:
            self._get_test_server_client().update_stored_campaigns()
        
    def __enter_current_scenario_log_file(self):
        if SessionContext.instance().log_manager.in_file and self.has_report_path:
            log_filename = self.current_scenario_report.get_path("logs", "report.log")
            SessionContext.instance().path_manager.makedirs(log_filename)
            # Note: do_remove_other_file_handlers is set to True, in case leaving previous scenario log file has failed.
            #       Normally this case shouldn't appear but it was already observed. And when it appeared, execution slowdowned drastically.
            SessionContext.instance().log_manager.enter_log_file(log_filename, do_remove_other_file_handlers=True)
        
    def __leave_current_scenario_log_file(self):
        if SessionContext.instance().log_manager.in_file and self.has_report_path:
            log_filename = self.current_scenario_report.get_path("logs", "report.log")
            SessionContext.instance().log_manager.leave_log_file(log_filename, do_remove_log_file=True)
    
    @classmethod
    def _get_test_server_client(cls):
        return SessionContext.instance().test_server_client
    
    @classmethod
    def _get_scenario_uid(cls, scenario):
        return f"{scenario.filename} at l.{scenario.line}"
    
    @classmethod
    def get_current_scenario_status_information(cls, scenario):
        scenario_uid = cls._get_scenario_uid(scenario)
        if scenario_uid not in cls._scenario_status_information_by_uid:
            step_failed, step_nb = cls.get_step_failed_info(scenario)
            
            # Define scenario status
            if step_failed is not None and hasattr(SessionContext.instance().get_scenario_context(), "is_in_preconditions") and SessionContext.instance().get_scenario_context().is_in_preconditions:
                status = "Failed in Preconditions"
            elif step_failed is not None and step_failed.keyword == "Given":
                status = "Failed in Given"
            elif step_failed is not None or scenario.status.has_failed():
                status = "Failed"
            else:
                status = scenario.status.name.capitalize()
            
            # Define scenario category
            category = None
            if cls._get_test_server_client().is_available:
                category = cls._compute_category_validation(scenario, status)
            
            cls._scenario_status_information_by_uid[scenario_uid] = [category, status, step_failed, step_nb, 
                                                                     SessionContext.instance().get_scenario_context(),
                                                                     SessionContext.instance().get_scenario_context().get_step(step_nb-1) if step_nb is not None else None]
        
        return cls._scenario_status_information_by_uid[scenario_uid]
    
    @classmethod
    def _compute_category_validation(cls, scenario, status):
        res = None
        
        # Get scenario execution statuses
        scenario_name = cls.format_scenario_short_description(scenario)
        sce_hist = cls._get_test_server_client().get_scenario_history(scenario_name=scenario_name, size=29)
        statuses = [s['status'] for s in reversed(sce_hist[0]['statuses'])] if sce_hist else []
        statuses.append(status)
        
        # Get scenario status sequences
        passed_sequences = []
        is_failed_relevant = None
        for status in statuses:
            if status == 'Passed':
                passed = True
            elif status.startswith("Failed"):
                passed = False
                if status == "Failed":
                    is_failed_relevant = True
                elif is_failed_relevant is None:
                    is_failed_relevant = False
            else:
                continue
            
            if len(passed_sequences) == 0 or passed != passed_sequences[-1][0]:
                passed_sequences.append([passed, 1])
            else:
                passed_sequences[-1][1] += 1
        
        # Compute category
        if passed_sequences:
            nb_exec = sum([x[1] for x in passed_sequences])
            last_passed, last_nb_times = passed_sequences[-1]
            if len(passed_sequences) == 1:
                if last_passed:
                    res = f'Always Success ({last_nb_times})'
                elif is_failed_relevant:
                    res = f'Always Failed ({last_nb_times})'
                else:
                    res = f'Always Not Relevant ({last_nb_times})'
            elif last_passed and len(passed_sequences) in [2, 3]:
                res = f'Fixed ({last_nb_times} success / {nb_exec})'
            elif len(passed_sequences) > 2:
                nb_fail = sum([x[1] for x in passed_sequences if not x[0]])
                if is_failed_relevant:
                    res = f'Random ({nb_fail} fails / {nb_exec})'
                else:
                    res = f'Random but Not Relevant ({nb_fail} fails / {nb_exec})'
            elif not last_passed:
                if is_failed_relevant:
                    res = f'Regression ({last_nb_times} fails / {nb_exec})'
                else:
                    res = f'Regression but Not Relevant ({last_nb_times} fails / {nb_exec})'
            else:
                res = f'Unknown (unmanaged sequence: {passed_sequences})'
        
        logger.debug(f"Category of scenario '{scenario}': {res}  (computed from last statuses: {statuses})")
        return res
        
    @classmethod
    def get_step_failed_info(cls, scenario):
        res_step, res_step_number = None, None
        for ind, step in enumerate(scenario.steps):
            if step.status.has_failed():
                res_step, res_step_number = step, ind+1
                break
        return res_step, res_step_number
    
    @classmethod
    def get_step_description(cls, step):
        res = "{} {}".format(step.keyword, step.name)
        text = cls.TStepTools.get_step_multiline_text(step, eval_params=EvaluateParameters.nothing(), raise_exception_if_none=False, log_level=logging.TRACE)  # @UndefinedVariable
        if text is not None:
            res += "\n\"\"\"\n{}\n\"\"\"".format(text) 
        if step.table:
            res += "\n{}".format(cls.TStepTools.represent_step_table(step.table, 4)) 
        return res
    
    @classmethod
    def get_step_error_message(cls, step):
        if step:
            if step.exception:
                return str(step.exception)
            elif step.error_message:
                return step.error_message
            elif step.status.is_undefined():
                return "Undefined step"
            elif step.status.is_pending():
                return "Step exists but is not implemented"
            elif step.status.has_failed():
                return "Unknown error (unexpected error case)"
        return None
    
    @classmethod
    def get_step_error(cls, step):
        if step:
            if step.exception:
                formatted_exception = Tools.represent_exception(step.exception)
                return "exception:\n{}".format(Tools.indent_string(4, formatted_exception))
            elif step.error_message:
                return "error_message:\n{}".format(Tools.indent_string(4, step.error_message))
            else:
                return cls.get_step_error_message(step)
        return None
    
    @classmethod
    def format_context_period(cls, context, format_precision_nsec=None, dt_ref=None, use_compact_format=default_value):
        dt_start, dt_end = context.start_datetime, context.end_datetime
        
        # Prepare format of start datetime
        if format_precision_nsec is not None:
            dt_start = DateTime.truncate_datetime(dt_start, precision_nanoseconds=format_precision_nsec)
        dt_format_start = cls._get_datetime_format_compared_to_reference(dt_start, dt_ref, use_compact_format=use_compact_format)
        
        # Prepare format of end datetime
        if dt_end is not None:
            if format_precision_nsec is not None:
                dt_end = DateTime.truncate_datetime(dt_end, precision_nanoseconds=format_precision_nsec)
            dt_format_end = cls._get_datetime_format_compared_to_reference(dt_end, dt_start, use_compact_format=use_compact_format)
            if len(dt_format_end) > len(dt_format_start):
                dt_format_start = dt_format_end
        else:
            dt_format_end = None
        
        # Format datetimes
        start_txt = DateTime.datetime_2_str(dt_start, dt_format=dt_format_start)
        end_txt = DateTime.datetime_2_str(dt_end, dt_format=dt_format_end) if dt_end is not None else ''
        
        # Truncate formatted datetimes if needed
        if format_precision_nsec is not None:
            trunc_len = len(f'{int(format_precision_nsec)}') - 4
            if trunc_len > 0:
                start_txt = start_txt[:-trunc_len-1]+'Z' if start_txt.endswith('Z') else start_txt[:-trunc_len]
                if len(end_txt) > 0:
                    end_txt = end_txt[:-trunc_len-1]+'Z' if end_txt.endswith('Z') else end_txt[:-trunc_len]
        
        return f"[{start_txt} - {end_txt}]"
    
    @classmethod
    def _get_datetime_format_compared_to_reference(cls, dt, dt_ref=None, use_compact_format=default_value):
        if use_compact_format is default_value:
            use_compact_format = Config.report_compact_datetime_period  # @UndefinedVariable
        
        if dt_ref is None:
            return FORMAT_DATETIME_ISO
        
        if not use_compact_format or dt.date() != dt_ref.date():
            return FORMAT_DATETIME_ISO
        elif dt.hour != dt_ref.hour:
            return FORMAT_TIME_ISO
        elif dt.minute != dt_ref.minute:
            return '%M:%S.%f'
        elif dt.second != dt_ref.second:
            return '%S.%f'
        else:
            return '.%f'
    
    @classmethod
    def format_scenario_short_description(cls, scenario):
        return f"{scenario.filename} at l.{scenario.line}"
    
    @classmethod
    def format_step_short_description(cls, step, step_number, step_context=None, dt_ref=None, has_failed=undefined_argument):
        if step:
            if step_context and step_context.start_datetime is not None:
                return f"step {step_number} (l.{step.line} on {cls.format_context_period(step_context, dt_ref=dt_ref)})"
            else:
                return f"step {step_number} (l.{step.line})"
        elif has_failed is not undefined_argument and has_failed:
            return "step ? (missing step implementation ?)"
        else:
            return None
            
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from typing import NamedTuple
from holado.common.context.session_context import SessionContext

logger = logging.getLogger(__name__)



class ExecutionHistoric():
    def __init__(self):
        self.__execution_historic = []
    
    def __iter__(self):
        return self.__execution_historic.__iter__()

    def __next__(self):
        return self.__execution_historic.__next__()

    def __has_scenario_context(self):
        return SessionContext.instance().has_scenario_context()
    
    @property
    def __multitask_manager(self):
        return SessionContext.instance().multitask_manager
    
    def before_all(self):
        pass
    
    def before_feature(self, feature_context, feature, feature_report=None):
        self.__execution_historic.append( self.__new_FeatureExecutionHistoric(feature_context, feature, feature_report) )
    
    def __new_FeatureExecutionHistoric(self, feature_context, feature, feature_report=None):
        res = NamedTuple("FeatureExecutionHistoric", feature_context=object, feature=object, feature_report=object, scenarios=list)
        res.feature_context = feature_context
        res.feature = feature
        res.feature_report = feature_report
        res.scenarios = []
        return res
    
    def before_scenario(self, scenario_context, scenario, scenario_report=None):
        seh = self.__new_ScenarioExecutionHistoric(scenario_context, scenario, scenario_report)
        self.__get_execution_historic_current_feature_scenarios().append(seh)
    
    def __new_ScenarioExecutionHistoric(self, scenario_context, scenario, scenario_report):
        res = NamedTuple("ScenarioExecutionHistoric", scenario_context=object, scenario=object, scenario_report=object, steps_by_scope=dict, 
                         category_validation=str, status_validation=str, step_failed=object, step_failed_number=int)
        res.scenario_context = scenario_context
        res.scenario = scenario
        res.scenario_report = scenario_report
        res.steps_by_scope = {}
        res.category_validation = None
        res.status_validation = None
        res.step_failed = None
        res.step_failed_number = None
        return res
    
    def before_step(self, step_context, step, step_level):
        seh = self.__new_StepExecutionHistoric(step_context, step)
        self.__get_execution_historic_current_scenario_steps(step_level).append(seh)
    
    def __new_StepExecutionHistoric(self, step_context, step, step_description=None):
        if step_description is None and step is not None:
            from holado_report.report.report_manager import ReportManager
            step_description = ReportManager.get_step_description(step)
        
        res = NamedTuple("StepExecutionHistoric", step_context=object, step=object, step_description=str, sub_steps=list)
        res.step_context = step_context
        res.step = step
        res.step_description = step_description
        res.sub_steps = []
        return res
    
    def after_step(self, step_context, step, step_level):
        pass
    
    def after_scenario(self, scenario, scenario_report=None):
        from holado_report.report.report_manager import ReportManager
        category_validation, status_validation, step_failed, step_number, _, _ = ReportManager.get_current_scenario_status_information(scenario)
        
        # Update execution historic
        current_scenario = self.__get_execution_historic_current_scenario()
        current_scenario.category_validation = category_validation
        current_scenario.status_validation = status_validation
        current_scenario.step_failed = step_failed
        current_scenario.step_failed_number = step_number
        
        # Prepare after scenario steps
        seh = self.__new_StepExecutionHistoric(step_context=None, step=None, step_description="After scenario steps")
        self.__get_execution_historic_current_scenario_steps(0).append(seh)
        
        
    def after_feature(self, feature, feature_report=None):
        pass
    
    def after_all(self):
        # logger.info(f"++++++++++ Execution historic size: {len(self.__execution_historic)}")
        pass
        
    def __get_execution_historic_current_feature_scenarios(self):
        if len(self.__execution_historic) > 0:
            return self.__execution_historic[-1].scenarios
        else:
            raise TechnicalException(f"No feature in execution historic")
        
    def __get_execution_historic_current_scenario(self):
        scenario_list = self.__get_execution_historic_current_feature_scenarios()
        if len(scenario_list) > 0:
            return scenario_list[-1]
        else:
            raise TechnicalException(f"No scenario in current feature in execution historic")
        
    def __get_execution_historic_current_scenario_steps(self, step_level):
        steps_by_scope = self.__get_execution_historic_current_scenario().steps_by_scope
        
        if self.__multitask_manager.is_main_thread:
            scope_name = "main"
        else:
            scope_name = self.__multitask_manager.get_thread_id()
        if scope_name not in steps_by_scope:
            steps_by_scope[scope_name] = []
        res = steps_by_scope[scope_name]
        
        for level in range(step_level):
            if len(res) == 0:
                logger.warning(f"Failed to get current scenario step list of level {step_level} (step list of level {level}: {res}) ; Add a fake step with description '[Missing step]'")
                res.append( self.__new_StepExecutionHistoric(None, None, "[Missing step]") )
            res = res[-1].sub_steps
        return res
    
    



import logging

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


class Config(object):
    """HolAdo project configuration"""
    
    # Application properties
    application_group = None
    
    # Timeouts
    session_timeout_seconds = 7 * 24 * 3600     # Session timeout is by default to 7 days
    timeout_seconds = 240            # Default timeout
    join_timeout_seconds = 1800      # Long timeout used when a join should stop without deadlock. When this timeout is reached, a TimeoutTechnicalException is raised

    # Message
    message_truncate_length = 500     # Length used to truncate a message when it is too long. When a message is truncated, it is suffixed by '...'

    # Time analysis
    threshold_warn_time_spent_s = 10    # Duration in seconds that triggers a warning on spent time

    # Redo
    redo_wait_process_min_interval_s = 0.001    # In redo, minimal duration to sleep between two verification of process thread end
    redo_wait_process_max_interval_s = 0.1      # In redo, maximal duration to sleep between two verification of process thread end

    # Unique values
    unique_string_padding_length = 6    # Number of characters used to generate a unique text suffix
     
    # Symbols
    DYNAMIC_SYMBOL = u"%"
    THREAD_DYNAMIC_SYMBOL = u"@"
    UNIQUE_SYMBOL = u"#"
    NOT_APPLICABLE_SYMBOL = u"N/A"
    NONE_SYMBOL = u"None"
    





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.tools.tools import Tools
import copy
from functools import total_ordering
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.standard_library.typing import Typing


@total_ordering
class TableCell(object):
    EMPTY_SYMBOL = u"None"
    # LEFT_MERGED_SYMBOL = u"LeftMerged";
    # TOP_MERGED_SYMBOL = u"TopMerged";
    
    """
    @summary: Table cell
    """
    def __init__(self, cell_content = None):
        self.__parent = None
        
        if cell_content is not None and isinstance(cell_content, str) and cell_content == TableCell.EMPTY_SYMBOL:
            self.__content = None
        else:
            self.__content = cell_content
    
    @property
    def parent(self):
        """ Parent, relevant for some table types. It's usually the origin of its creation. """
        return self.__parent
    
    @parent.setter
    def parent(self, parent):
        """ Set parent. """
        self.__parent = parent

    @property
    def content(self):
        return self.__content
    
    @content.setter
    def content(self, cell_content):
        self.__content = cell_content
    
    @property
    def string_content(self):
        if self.__content is not None:
            return str(self.__content)
        else:
            return None
    
    def _verify_valid_compared_object(self, other, raise_exception=True):
        res = isinstance(other, TableCell)
        if not res and raise_exception:
            return TechnicalException(f"Unmanaged comparison between types '{self.__class__.__name__}' and '{other.__class__.__name__}'")
        
    def __eq__(self, other):
        self._verify_valid_compared_object(other)
        return (self.content == other.content)
    
    def __lt__(self, other):
        self._verify_valid_compared_object(other)
        return (self.content < other.content)

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    def __eq__(self, other):
        if not isinstance(other, TableCell):
            return NotImplemented(f"Not implemented to compare a {Typing.get_object_class_fullname(self)} to a {Typing.get_object_class_fullname(other)}")
        return (self.content == other.content)
    
    def __lt__(self, other):
        if not isinstance(other, TableCell):
            return NotImplemented(f"Not implemented to compare a {Typing.get_object_class_fullname(self)} to a {Typing.get_object_class_fullname(other)}")
        return (self.content < other.content)
    
    def __repr__(self)->str:
        return f"<{self.__class__.__name__}>[{self.represent()}]"
        
    def is_empty(self, **kwargs):  # @UnusedVariable
        return (self.__content is None)
        
    def represent(self, indent = 0, cell_value_prefix = "'", cell_value_postfix = "'"):
        res_list = []
        
        res_list.append(Tools.get_indent_string(indent))
        
        if self.is_empty():
            res_list.append(TableCell.EMPTY_SYMBOL)
        elif isinstance(self.__content, str):
            if cell_value_prefix is not None:
                res_list.append(cell_value_prefix)
            res_list.append(self.string_content)
            if cell_value_postfix is not None:
                res_list.append(cell_value_postfix)
        else:
            res_list.append(self.string_content)
        
        return "".join(res_list)
        
    def compare_to(self, cell):
        raise NotImplementedError()
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################



from builtins import super
from holado_core.common.tables.table_row import TableRow
import copy
from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing
from holado_python.common.iterables import is_sorted

logger = logging.getLogger(__name__)


class Table(object):
    def __init__(self):
        super().__init__()
        
        self.__content = None
        self.__parent = None
        
        self.__rows = []
        
    def __iter__(self):
        return self.__rows.__iter__()

    def __next__(self):
        return self.__rows.__next__()
    
    def __getitem__(self, key):
        if isinstance(key, int):
            return self.__rows[key]
        else:
            raise TechnicalException(f"Unmanaged key: {key} (type: {Typing.get_object_class_fullname(key)})")
    
    def __copy__(self):
        cls = self.__class__
        res = cls()
        
        # Copy rows
        for row in self.rows:
            res.add_row(row=copy.copy(row))
        
        return res

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result
    
    def __repr__(self)->str:
        if self.nb_rows < 10 or self.nb_rows * self.nb_columns < 100:
            return f"<{self.__class__.__name__}>({self.nb_rows}x{self.nb_columns})\n{self.represent()}"
        else:
            return f"<{self.__class__.__name__}>({self.nb_rows}x{self.nb_columns})\n{self.represent(limit_rows=10)}[...]"
    
    @property
    def content(self):
        """ Content of whole table, relevant for some table types. """
        return self.__content
    
    @content.setter
    def content(self, content):
        """ Set content of whole table, relevant for some table types. """
        self.__content = content
    
    @property
    def parent(self):
        """ Parent, relevant for some table types. It's usually the origin of its creation. """
        return self.__parent
    
    @parent.setter
    def parent(self, parent):
        """ Set parent. """
        self.__parent = parent
    
    @property
    def rows(self):
        return self.__rows
    
    @property
    def nb_columns(self):
        if self.nb_rows > 0:
            return len(self.rows[0])
        else:
            return 0
    
    @property
    def nb_rows(self):
        return len(self.__rows)
            
    def add_row(self, row=None, cells=None, cells_content=None, **kwargs):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"add_row({row=}, {cells=}, {cells_content=})")
        if row is not None:
            self.__rows.append(row)
            return row
        else:
            return self.add_row(row=self._new_row(cells=cells, cells_content=cells_content, **kwargs), **kwargs)
        
    def extend(self, table, copy_method=None, **kwargs):  # @UnusedVariable
        for row in table.rows:
            if copy_method is not None:
                new_row = copy_method(row)
            else:
                new_row = row
            self.add_row(row=new_row, **kwargs)
        
    def _new_row(self, cells=None, cells_content=None, **kwargs):  # @UnusedVariable
        return TableRow(cells=cells, cells_content=cells_content)
        
    def get_row(self, index, **kwargs) -> TableRow :  # @UnusedVariable
        return self.rows[index]
    
    def get_column(self, index, **kwargs):
        res = self._new_row(**kwargs)
        for row in self.rows:
            res.add_cell(cell=row[index], **kwargs)
        return res

    def add_column(self, col=None, cells=None, cells_content=None, **kwargs):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"add_column({col=}, {cells=}, {cells_content=})")
        if col is not None:
            self.add_column(cells=col.cells, **kwargs)
        elif cells is not None:
            if len(cells) != self.nb_rows:
                raise TechnicalException(f"Length of cells is {len(cells)} whereas table has {self.nb_rows} rows")
            for i, cell in enumerate(cells):
                self.get_row(i).add_cell(cell=cell, **kwargs)
        elif cells_content is not None:
            if len(cells_content) != self.nb_rows:
                raise TechnicalException(f"Length of cells is {len(cells_content)} whereas table has {self.nb_rows} rows")
            for i, cell_content in enumerate(cells_content):
                self.get_row(i).add_cell(content=cell_content, **kwargs)
        else:
            for row in self.rows:
                row.add_cell(content=None, **kwargs)
        
    def remove_column(self, index=None, **kwargs):
        if index is None:
            raise TechnicalException("index must be specified")
        
        for row in self.rows:
            row.remove_cell(index, **kwargs)
        
    def remove_row(self, index=None, **kwargs):  # @UnusedVariable
        if index is None:
            raise TechnicalException("index must be specified")
        elif index >= self.nb_rows:
            raise FunctionalException(f"index is out of range ({index} >= {self.nb_rows})")
        
        return self.rows.pop(index)
    
    def represent(self, indent = 0, limit_rows=-1):
        res_list = []
        has_limited = False
        for row in self.rows:
            if limit_rows >=0 and len(res_list) >= limit_rows:
                has_limited = True
                break
            
            if row is not None:
                res_row = row.represent(indent)
            else:
                res_row = "{ROW_IS_NONE}"
            res_row += "\n"
            res_list.append(res_row)
            
        res = "".join(res_list)
        if has_limited:
            res += Tools.indent_string(indent, "[...]\n")
        return res
    
    def switch_columns(self, index_1, index_2, **kwargs):
        if index_1 == index_2:
            raise TechnicalException(f"The two indexes are equal (index = {index_1})")
    
        for row in self.rows:
            row.switch_cells(index_1, index_2, **kwargs)
            
    def sort(self, indexes, reverse=False, **kwargs):  # @UnusedVariable
        """Sort table rows according to indexes order."""
        if not indexes:
            raise TechnicalException("At least one index must be defined")
        self.__rows.sort(key=lambda x: tuple(x[i] for i in indexes), reverse=reverse)
    
    def is_sorted(self, indexes, reverse=False, **kwargs):  # @UnusedVariable
        """Check if table rows are sorted according to indexes order."""
        if not indexes:
            raise TechnicalException("At least one index must be defined")
        return is_sorted(self.__rows, key=lambda x: tuple(x[i] for i in indexes), reverse=reverse)

    def order_columns(self, indexes=None, **kwargs):
        orig_col_indexes = list(range(self.nb_columns))
        for col_index, orig_col_ind in enumerate(indexes):
            cur_col_ind = orig_col_indexes.index(orig_col_ind)
            if cur_col_ind != col_index:
                self.switch_columns(index_1=col_index, index_2=cur_col_ind, **kwargs)
                orig_col_indexes[col_index], orig_col_indexes[cur_col_ind] = orig_col_indexes[cur_col_ind], orig_col_indexes[col_index]
        
    def remove_rows_verifying(self, expected_table, table_comparator, keep_rows=False, **kwargs):
        """
        Remove rows that are equals compared to at least one line of expected_table.
        If keep_rows is True, keep the rows rather than remove them. 
        """
        n = 0
        while n < self.nb_rows:
            row = self.get_row(n, **kwargs)
            
            found = False
            for i in range(expected_table.nb_rows):
                found = table_comparator.row_comparator.equals(row, expected_table.get_row(i), raise_exception=False);
                if found:
                    break
            
            if found ^ keep_rows:
                self.remove_row(n)
            else:
                n += 1
            
    def remove_rows_duplicated(self, row_comparator, **kwargs):
        """
        Remove rows that are duplicates of previous rows, using given row_comparator.
        """
        n = 1
        while n < self.nb_rows:
            row = self.get_row(n)
            
            found = False
            for i in range(n):
                found = row_comparator.equals(row, self.get_row(i), raise_exception=False);
                if found:
                    break
            
            if found:
                self.remove_row(n, **kwargs)
            else:
                n += 1
            
    def is_empty(self, **kwargs):  # @UnusedVariable
        return self.nb_rows == 0





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from timeit import default_timer as timer
from holado_core.common.finders.finder import Finder
import logging

logger = logging.getLogger(__name__)


class ElementFinder(Finder):
    """ Generic finder of elements.
    """
    
    def __init__(self, description=None):
        super().__init__(description)
    
    def _build_result_element(self, res, find_context, find_parameters, element_description=None):
        super()._build_result_element(res, find_context, find_parameters)
        
        if element_description is not None and len(element_description) > 0:
            res.description = element_description
        elif self.element_description is not None:
            res.description = self.element_description
            
        return res
    
    def _build_result_list(self, res, find_context, find_parameters, element_description=None):
        for el in res:
            self._build_result_element(el, find_context, find_parameters, element_description)
        return res
    
    def is_valid_input(self, element, find_context, find_parameters):
        return self.is_valid_element(element, find_context, find_parameters)
    
    def is_valid_output(self, element, find_context, find_parameters):
        return self.is_valid_element(element, find_context, find_parameters)
    
    def is_valid_element(self, element, find_context, find_parameters):
        """
        Return if element is valid according this finder.
        @param element Element to validate
        @param find_context Find context
        @param find_parameters Find parameters
        @return True if element is valid according this finder
        """
        raise NotImplementedError(f"Unimplemented is_valid_element since inconsistent for [{self.finder_description}]")
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
from holado_scripting.text.interpreter.text_interpreter import TextInterpreter
from holado_scripting.common.tools.dynamic_text_manager import DynamicTextManager
from holado_scripting.common.tools.variable_manager import VariableManager
from holado.common.context.context import Context
import logging
from holado_scripting.common.tools.expression_evaluator import ExpressionEvaluator
from holado.common.context.session_context import SessionContext
from holado_core.common.block.scope_manager import ScopeManager
from holado_core.common.block.block_manager import BlockManager
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_scripting.text.verifier.text_verifier import TextVerifier
from holado_multitask.multitasking.multitask_manager import MultitaskManager
from holado_python.common.tools.datetime import DateTime

logger = logging.getLogger(__name__)


class ScenarioContext(Context):
    def __init__(self, scenario):
        super().__init__("Scenario", with_post_process=True)
        
        self.__scenario = scenario
        
        self.__start_date = DateTime.now()
        self.__end_date = None
        
        self.__main_thread_uid = SessionContext.instance().multitask_manager.main_thread_uid
        self.__steps_by_thread_uid = {}
        
    def __str__(self):
        return f"{{ScenarioContext({id(self)}):{self.scenario.name}}}"

    @property
    def scenario(self):
        return self.__scenario
    
    @property
    def start_datetime(self):
        return self.__start_date
    
    @property
    def end_datetime(self):
        return self.__end_date
    
    @property
    def duration(self):
        if self.__end_date is not None:
            return (self.__end_date - self.__start_date).total_seconds()
        else:
            return None
    
    def get_nb_steps(self, thread_uid=None):
        if thread_uid is None:
            thread_uid = MultitaskManager.get_thread_uid()
        if thread_uid in self.__steps_by_thread_uid:
            return len(self.__steps_by_thread_uid[thread_uid])
        else:
            return 0
        
    def has_step(self, thread_uid=None):
        if thread_uid is None:
            thread_uid = MultitaskManager.get_thread_uid()
        return thread_uid in self.__steps_by_thread_uid and len(self.__steps_by_thread_uid[thread_uid]) > 0
    
    def get_current_step(self, thread_uid=None):
        return self.get_step(-1, thread_uid=thread_uid)
    
    def add_step(self, step_context, thread_uid=None):
        if thread_uid is None:
            thread_uid = MultitaskManager.get_thread_uid()
        if thread_uid not in self.__steps_by_thread_uid:
            self.__steps_by_thread_uid[thread_uid] = []
        self.__steps_by_thread_uid[thread_uid].append(step_context)
    
    def get_step(self, index, thread_uid=None):
        if thread_uid is None:
            thread_uid = MultitaskManager.get_thread_uid()
        if not self.has_step(thread_uid):
            raise TechnicalException(f"Scenario has no step (for thread UID '{thread_uid}'")
        if index >= len(self.__steps_by_thread_uid[thread_uid]):
            raise TechnicalException(f"Step index exceeds number of steps in thread of UID '{thread_uid}' (index {index} >= {len(self.__steps_by_thread_uid[thread_uid])})")
        return self.__steps_by_thread_uid[thread_uid][index]
    
    @property
    def block_manager(self):
        if not self.has_object("block_manager"):
            self.set_object("block_manager", BlockManager())
        return self.get_object("block_manager")
    
    @property
    def scope_manager(self):
        if not self.has_object("scope_manager"):
            self.set_object("scope_manager", ScopeManager())
        return self.get_object("scope_manager")


    def has_dynamic_text_manager(self):
        return self.has_object("dynamic_text_manager")
        
    def get_dynamic_text_manager(self) -> DynamicTextManager:
        if not self.has_dynamic_text_manager():
            dynamic_text_manager = DynamicTextManager("scenario")
            self.set_object("dynamic_text_manager", dynamic_text_manager)
            dynamic_text_manager.initialize(SessionContext.instance().unique_value_manager)
        return self.get_object("dynamic_text_manager")
        
    def has_text_interpreter(self):
        return self.has_object("text_interpreter")
        
    def get_text_interpreter(self):
        if not self.has_text_interpreter():
            interpreter = TextInterpreter()
            self.set_object("text_interpreter", interpreter)
            interpreter.initialize(self.get_variable_manager(), self.get_expression_evaluator(), self.get_text_verifier(), self.get_dynamic_text_manager())
        return self.get_object("text_interpreter")
        
    def has_text_verifier(self):
        return self.has_object("text_verifier")
        
    def get_text_verifier(self):
        if not self.has_text_verifier():
            verifier = TextVerifier()
            self.set_object("text_verifier", verifier)
            verifier.initialize(self.get_variable_manager(), self.get_expression_evaluator(), self.get_text_interpreter())
        return self.get_object("text_verifier")
        
    def has_variable_manager(self):
        return self.has_object("variable_manager")
        
    def get_variable_manager(self) -> VariableManager:
        if not self.has_variable_manager():
            manager = VariableManager(SessionContext.instance().get_feature_context().get_variable_manager())
            self.set_object("variable_manager", manager)
            file_path = SessionContext.instance().report_manager.current_scenario_report.get_path("logs", "variable_update.log") if SessionContext.instance().with_session_path else None
            manager.initialize(self.get_dynamic_text_manager(), SessionContext.instance().unique_value_manager,
                               variable_update_log_file_path=file_path)
        return self.get_object("variable_manager")
    
    def has_expression_evaluator(self):
        return self.has_object("expression_evaluator")
    
    def get_expression_evaluator(self) -> ExpressionEvaluator:
        if not self.has_expression_evaluator():
            evaluator = ExpressionEvaluator()
            self.set_object("expression_evaluator", evaluator)
            uvm = SessionContext.instance().unique_value_manager
            evaluator.initialize(self.get_dynamic_text_manager(), uvm, self.get_text_interpreter(), self.get_variable_manager())
        return self.get_object("expression_evaluator")
    
    def end(self):
        # End scenario
        self.__end_date = DateTime.now()


        


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.criterias.tools.criteria_context import CriteriaContext
from holado_core.common.criterias.tools.criteria_parameters import CriteriaParameters
import logging
from holado_ui.ui.gui.criterias.enums import CheckEditableModes
from holado_ui.ui.gui.criterias.gui_criteria import GUICriteria
from holado_core.common.tools.tools import Tools
from holado.holado_config import Config
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_scripting.text.verifier.text_verifier import TextVerifier
from holado_ui_selenium.ui.gui.selenium.criterias.selenium_criteria import SeleniumCriteria
from holado_core.common.exceptions.functional_exception import FunctionalException

logger = logging.getLogger(__name__)



class SeleniumCriteriaAndVerifier(GUICriteria):
    """ Apply a SeleniumCriteria, and if validated apply verifications by verifier.
    """
    
    def __init__(self, criteria:SeleniumCriteria, text_verifier:TextVerifier):
        super().__init__(CheckEditableModes.NoCheck)
        
        self.__criteria = criteria
        self.__text_verifier = text_verifier
        
        self.__attribute_value_by_attribute_name = {}
    
    def get_attribute(self, attr_name):
        """
        @param attr_name Attribute name
        @return Attribute value
        """
        return self.__attribute_value_by_attribute_name.get(attr_name)
    
    def set_attribute(self, attr_name, attr_value):
        """
        @param attr_name Attribute name
        @param attr_value Attribute value
        """
        if attr_value is not None and len(attr_value) > 0:
            self.__attribute_value_by_attribute_name[attr_name] = attr_value
        elif attr_name in self.__attribute_value_by_attribute_name:
            del self.__attribute_value_by_attribute_name[attr_name]
    
    
    def validate(self, element, criteria_context: CriteriaContext, criteria_parameters: CriteriaParameters):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"{self._get_ident_string_level(criteria_parameters)}[{self} ; [{self._internal_api.get_element_description(element)}]] validate : begin")
        
        # Prepare analyze time spent
        if criteria_parameters.analyze_time_spent:
            start = Tools.timer_s()
        
        # Validate element
        res = self.criteria.validate(element, criteria_context, criteria_parameters)

        # If validated, verify
        if res:
            try:
                res = self.__check_element(element)
            except FunctionalException as e:
                raise TechnicalException(str(e)) from e
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"{self._get_indent_string_level(criteria_parameters)}[{self} ; [{self._internal_api.get_element_description(element)}]] validate : end -> {res}")
        
        # Analyze time spent
        if criteria_parameters.analyze_time_spent:
            duration = Tools.timer_s() - start
            if duration > Config.threshold_warn_time_spent_s:
                logger.warning(f"{self._get_indent_string_level(criteria_parameters)}[{self} ; [{self._internal_api.get_element_description(element)}]] validate -> {res} (took {duration} s)")
        
        return res
    
    def __check_element(self, element):
        res = True
        for attr_name, attr_value in self.__attribute_value_by_attribute_name.items():
            web_element = element.element
            value = web_element.get_attribute(attr_name)
            if value is not None:
                res = self.__text_verifier.check(attr_value, value)
            else:
                res = False
            if not res:
                break

        return res;
    
    def __str__(self):
        res_list = [super().__str__()]
        
        res_list.append(f"&criteria=[{self.__criteria}]")

        if not self.__attribute_value_by_attribute_name.empty():
            res_list.append("&attributes={")
            attr_list = []
            for key, value in self.__attribute_value_by_attribute_name.items():
                attr_list.append(f"'{key}':'{value}'")
            res_list.append(", ".join(attr_list))
            res_list.append("}")
        
        return "".join(res_list)
    
    
    


from django.contrib.auth.models import User
from django_grpc_framework import proto_serializers
import account_pb2


class UserProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = User
        proto_class = account_pb2.User
        fields = ['id', 'username', 'email', 'groups']
        




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.finders.tools.find_builder import FindBuilder
from holado_core.common.inspectors.tools.inspect_builder import InspectBuilder
from holado_core.common.inspectors.tools.inspect_parameters import InspectParameters
from holado_core.common.inspectors.tools.inspect_context import InspectContext
import abc
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)



class Inspector(object):
    """ Base class for inspector.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, module_name):
        # Modules management
        self.__module_name = module_name
        self.__inspector_by_module_name = {}
    
        # Builders to use
        self.__inspect_builder = None
        self.__find_builder = None
    
    
    def initialize(self):
        """
        Initialize inspector
        """
        self.__inspect_builder = self.default_inspect_builder()
        self.__find_builder = self.default_find_builder()
    
    @property
    def find_builder(self):
        """
        @return Find builder defined for this inspector
        """
        return self.__find_builder
    
    @property
    def inspect_builder(self):
        """
        @return Inspect builder defined for this inspector
        """
        return self.__inspect_builder
    
    @property
    def module_name(self):
        return self.__module_name
    
    @property
    def default_inspect_builder(self):
        """
        @return New default inspect builder for this inspector
        """
        res = InspectBuilder()
        res.initialize(self)
        return res
    
    @property
    def default_find_builder(self):
        """
        @return New default find builder for this inspector
        """
        return FindBuilder()
    
    def _get_inspector_for_module(self, name):
        """
        @param name Module name
        @return Module inspector
        """
        if name not in self.__inspector_by_module_name:
            self.__inspector_by_module_name[name] = self._initialize_module(name)
        return self.__inspector_by_module_name[name]
    
    def _initialize_module(self, name):
        """
        Initialize a new instance of given module name
        @param name Module name
        @return New module instance
        """
        raise TechnicalException(f"Unmanaged module '{name}'")
    
    def _get_finder_from_modules(self, method_name, method_args=None, method_kwargs=None, inspect_context:InspectContext=None, inspect_parameters:InspectParameters=None):
        method_args = [] if method_args is None else method_args
        method_kwargs = {} if method_kwargs is None else method_kwargs
        inspect_context = self.inspect_builder.context() if inspect_context is None else inspect_context
        inspect_parameters = self.inspect_builder.parameters() if inspect_parameters is None else inspect_parameters
        
        finders = []
        for module_name in self.inspect_builder.module_names:
            module = self._get_inspector_for_module(module_name)
            if hasattr(module, method_name):
                method = getattr(module, method_name)
                module_parameters = self.inspect_builder.get_parameters_for_module(module_name, inspect_parameters)
                try:
                    finder = method(*method_args, **method_kwargs, inspect_context=inspect_context, inspect_parameters=module_parameters)
                except FunctionalException as e:
                    raise TechnicalException("Unexpected exception") from e
                
                if finder is not None:
                    self._add_finder_to_list(finders, finder, inspect_context, inspect_parameters)
        
        return self._build_result_finder_from_list(finders=finders, inspect_context=inspect_context, inspect_parameters=inspect_parameters)
    
    def _remove_none_from_list(self, list_):
        if list_ is None:
            return
        
        i = 0
        while i < len(list_):
            if list_[i] is None:
                del list_[i]
            else:
                i += 1
    
    def _build_result_finder(self, finder, inspect_context:InspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Build result finder, considering given inspect parameters.
        @param finder Finder to return
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Result finder
        """
        if finder is None:
            if inspect_parameters.raise_no_such_finder:
                raise TechnicalException(f"No such finder")
            else:
                return None
        
        if inspect_context.finder_type is not None and inspect_context.finder_type not in inspect_parameters.finder_types:
            return None
        
        finder.inspector = self
        if self.__module_name is not None:
            finder.info.module_name = self.__module_name
        if inspect_context.finder_type is not None:
            finder.info.finder_type = inspect_context.finder_type
        
        return finder
    
    
    def _build_result_finder_from_list(self, finders, inspect_context:InspectContext=None, inspect_parameters:InspectParameters=None, element_description=None):
        """
        Build result finder with description copied from first finder in list
        @param finders Finders list
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @param element_description Element description
        @return Result finder
        """
        if element_description is None and finders is not None and len(finders) > 0:
            element_description = finders[0].element_description
        
        self._remove_none_from_list(finders)
    
        if finders is None or len(finders) == 0:
            if inspect_parameters.raise_no_such_finder:
                raise TechnicalException(f"No such finder")
            else:
                return None
        elif finders.size() == 1:
            return self._build_result_finder(finders[0], inspect_context, inspect_parameters)
        else:
            res = inspect_parameters.result_finder_type(element_description)
            for finder in finders:
                if hasattr(res, "set_next_finder"):
                    res.set_next_finder(finder)
                elif hasattr(res, "add_finder"):
                    res.add_finder(finder)
                else:
                    raise TechnicalException(f"Unmanaged result finder type '{Typing.get_object_class_fullname(res)}'")
            return self._build_result_finder(res, inspect_context, inspect_parameters)
        
    def _merge_result_finders(self, result_finders, inspect_context:InspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Merge result finders to first result finder
        @param result_finders Finders list
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Result finder
        """
        self._remove_none_from_list(result_finders)
        
        if result_finders is None or len(result_finders) == 0:
            if inspect_parameters.raise_no_such_finder:
                raise TechnicalException(f"No such finder")
            else:
                return None
        else:
            res = result_finders[0]
            for result_finder in result_finders[1:]:
                if hasattr(res, "set_next_finder"):
                    res.set_next_finder(result_finder)
                elif hasattr(res, "add_finder"):
                    res.add_finder(result_finder)
                else:
                    raise TechnicalException(f"Unmanaged result finder type '{Typing.get_object_class_fullname(res)}'")
            return res
    
    def _add_finder_to_list(self, finders, finder, inspect_context:InspectContext=None, inspect_parameters:InspectParameters=None):
        if finder is not None:
            finders.add(self._build_result_finder(finder, inspect_context, inspect_parameters))
    
    


# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import behave

from holado.common.context.session_context import SessionContext
from holado_core.common.tools.tools import Tools
import types
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_test.scenario.step_tools import StepTools
from holado_test.test_config import TestConfig
import time
from behave.model_core import Status


logger = logging.getLogger(__name__)


# Use regular expression in step definitions
behave.use_step_matcher(BehaveStepTools.DEFAULT_STEP_MATCHER)


def __has_scenario_context():
    return SessionContext.instance().has_scenario_context()

def __get_scope_manager():
    if __has_scenario_context():
        return SessionContext.instance().get_scenario_context().scope_manager
    else:
        return None


def before_all(context):
    try:
        SessionContext.instance().before_all(context)
    except:
        logger.exception("Hook error")
        raise

def after_all(context):
    try:
        SessionContext.instance().after_all()
    except:
        logger.exception("Hook error")
        raise
    
def before_feature(context, feature):
    try:
        SessionContext.instance().before_feature(feature)
    except:
        logger.exception(f"Hook error (feature: {feature})")
        raise

def after_feature(context, feature):
    try:
        SessionContext.instance().after_feature(feature)
    except:
        logger.exception(f"Hook error (feature: {feature})")
        raise

def before_scenario(context, scenario):
    try:
        SessionContext.instance().before_scenario(scenario)
    except:
        logger.exception(f"Hook error (scenario: {scenario})")
        raise

def after_scenario(context, scenario):
    # Process after step for undefined and skipped steps
    try:
        for step in scenario.steps[SessionContext.instance().get_scenario_context().get_nb_steps():]:
            after_step(context, step, has_started=False)
    except:
        logger.exception(f"Hook error (scenario: {scenario})")
    
    # Process after scenario
    try:
        SessionContext.instance().after_scenario(scenario)
    except:
        logger.exception(f"Hook error (scenario: {scenario})")
        raise

def before_step(context, step):
    try:
        scope_manager = __get_scope_manager()
        logger.info(f"Step [{__step_description(step)}] (level: {scope_manager.scope_level('steps') if scope_manager else '?'})")
        
        SessionContext.instance().before_step(step)
    except:
        # logger.exception(f"Hook error (step: {step})")
        logger.exception(f"Hook error (step: {step})", stack_info=True)
        raise
    
    context.step = step

def after_step(context, step, has_started=True):
    try:
        # Update step if needed (ex: if step raised an expected exception, the step is in fact passing)
        if hasattr(context, "do_update_step_status"):
            # logger.warning(f"+++++++++++ Update step status to {context.do_update_step_status}")
            step.status = context.do_update_step_status
            del context.do_update_step_status
        
        # Session context
        SessionContext.instance().after_step(step, has_started=has_started)
        
        # Log step result
        step_descr = __step_description(step)
        try:
            var_step_descr = StepTools.evaluate_variable_name(step_descr, log_level=logging.NOTSET)
        except:
            var_step_descr = "[ERROR during evaluation]"
        try:
            interpreted_step_descr = StepTools.evaluate_string_parameter(step_descr, log_level=logging.NOTSET)
        except:
            interpreted_step_descr = "[ERROR during interpretation]"
        if step.status in [Status.failed, Status.error] or interpreted_step_descr != step_descr or var_step_descr != interpreted_step_descr:
            msg = "Step {}:\n    step:\n{}{}{}".format(
                step.status.name, Tools.indent_string(8, step_descr),
                "" if interpreted_step_descr == step_descr else "\n    interpreted step:\n{}".format(Tools.indent_string(8, interpreted_step_descr)),
                "" if var_step_descr == interpreted_step_descr else "\n    partially interpreted step:\n{}".format(Tools.indent_string(8, var_step_descr)) )
            
            if step.status in [Status.failed, Status.error]:
                msg += "\n    error:\n{}\n    step attributes:\n{}".format(
                    Tools.indent_string(8, __step_error(step)),
                    Tools.indent_string(8, __step_attributes(step)) )
                if step.hook_failed:
                    msg += "\n    Hook failed"
        elif '\n' in step_descr:
            msg = "Step {}:\n{}".format(step.status.name, Tools.indent_string(4, step_descr))
        else:
            msg = "Step {}: {}".format(step.status.name, step_descr)
        
        if step.status in [Status.failed, Status.error]:
            logger.error(msg)
            
            if TestConfig.wait_on_step_failure_s > 0:
                time.sleep(TestConfig.wait_on_step_failure_s)
        else:
            logger.debug(msg)
    except:
        logger.exception(f"Hook error (step: {step})")
        raise

def __step_description(step):
    return SessionContext.instance().report_manager.get_step_description(step)

def __step_error(step):
    return SessionContext.instance().report_manager.get_step_error(step)

def __step_attributes(step):
    res_list = []
    for attr_name in dir(step):
        if not attr_name.startswith("__") and type(getattr(step, attr_name)) not in [types.MethodType, types.MethodWrapperType]:
            res_list.append("{}: {}".format(attr_name, str(getattr(step, attr_name))))
    return "\n".join(res_list)
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_core.common.finders.tools.find_context import ContainerFindContext
from holado_core.common.exceptions.element_exception import NoSuchElementException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.drivers.element_internal_api import ElementInternalAPI
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class TreeInternalAPI(ElementInternalAPI):
    """ Base class for elements tree internal API.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, driver):
        super().__init__(driver)
    
    def find_children(self, criteria, container, find_context, find_parameters):
        """
        Find the children elements verifying given criteria.
        @param container Container in which search
        @param criteria Criteria to verify
        @param find_context Find context
        @param find_parameters Find parameters
        @return Children elements
        """
        if container is None:
            if isinstance(find_context, ContainerFindContext) and find_context.container is not None:
                container = find_context.container
            else:
                raise TechnicalException("Find context has no container")
            
        children = self.find_children_elements(container, find_context, find_parameters)
        return self.validate_elements(criteria, children, find_context, find_parameters)
    
    def find_children_elements(self, container, find_context, find_parameters):
        """
        Find the children elements of given element.
        @param container An element
        @param find_context Find context
        @param find_parameters Find parameters
        @return Children elements
        """
        raise NotImplementedError

    def find_element_in_tree(self, criteria, container, find_context, find_parameters):
        """
        Find, in container tree, the element verifying given criteria.
        @param container Container in which search
        @param criteria Criteria to verify
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found element
        """
        sub_find_parameters = find_parameters.with_analyze_time_spent(False)
        candidates = self.find_elements_in_tree(criteria, container, find_context, sub_find_parameters)
        return self._get_element_from_list(candidates, 1, find_context, sub_find_parameters)
    
    def find_elements_in_tree(self, criteria, container, find_context, find_parameters):
        """
        Find, in container tree, all elements verifying given criteria.
        When an element verifies given criteria, its sub tree is not explored.
        
        @param container Container in which search
        @param criteria Criteria to verify
        @param find_context Find context
        @param find_parameters Find parameters
        @return List of elements
        """
        if container is None:
            if isinstance(find_context, ContainerFindContext) and find_context.container is not None:
                container = find_context.container
            else:
                raise TechnicalException("Find context has no container")
            
        res = []
        sub_find_parameters = find_parameters.get_next_level().with_valid_container(True)
        criteria_context = find_context.get_criteria_context()
        criteria_parameters = find_parameters.get_criteria_parameters()
        
        # Verify this element
        is_validated = False
        if find_parameters.add_valid_container:
            is_validated = criteria.validate(container, criteria_context, criteria_parameters)
            if is_validated:
                res.append(container)
        
        # Verify in its children
        if not is_validated:
            children = self.find_children(None, container, find_context, sub_find_parameters.with_nb_max_elements(None))
            for child in children:
                if find_parameters.nb_max_elements is not None:
                    # Search only remaining max elements to find
                    res.extend( self.find_elements_in_tree(criteria, child, find_context, sub_find_parameters.with_nb_max_elements(find_parameters.nb_max_elements - len(res))) )
                       
                    # Stop if max number of elements is reached
                    if len(res) >= find_parameters.nb_max_elements:
                        break
                else:
                    res.extend( self.find_elements_in_tree(criteria, child, find_context, sub_find_parameters) )
            
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"find_elements_in_tree: criteria={criteria} ; container={container} ; max nb={find_parameters.nb_max_elements} ; visibility={find_parameters.visibility}     ->  {len(res)} elements")
        return res
    
    def find_parent(self, criteria, element, find_context, find_parameters):
        """
        Find the parent element for given element, verifying given criteria.
        @param element An element
        @param criteria Criteria
        @param find_context Find context
        @param find_parameters Find parameters
        @return The parent element
        """
        if element is None:
            if isinstance(find_context, ContainerFindContext) and find_context.container is not None:
                element = find_context.container
            else:
                raise TechnicalException("Find context has no container")
            
        parent_element = element
        found = False
        sub_find_parameters = find_parameters.get_next_level()
        criteria_context = find_context.get_criteria_context()
        criteria_parameters = sub_find_parameters.get_criteria_parameters()
        
        while not found:
            # Find parent
            parent_element = self.find_parent_element(parent_element, find_context, sub_find_parameters)
            
            # Break if reached top of hierarchy
            if parent_element is None:
                break

            # Verify criteria
            if criteria is not None:
                found = criteria.validate(parent_element, criteria_context, criteria_parameters)
            else:
                found = True
            
        if not found:
            if find_parameters.raise_no_such_element:
                raise NoSuchElementException(f"Unable to find parent with criteria: {criteria}")
            else:
                parent_element = None
           
        return parent_element
    
    def find_parent_element(self, element, find_context, find_parameters):
        """
        Find the parent element for given element.
        @param element An element
        @param find_context Find context
        @param find_parameters Find parameters
        @return The parent element
        """
        raise NotImplementedError
    
    
    
    


import os
import sys
import argparse
import logging

here = os.path.abspath(os.path.dirname(__file__))

holado_path = os.path.join(here, "..", "..")
sys.path.insert(0, os.path.join(holado_path, "src"))


from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_report.report.analyze.scenario_duration_manager import ScenarioDurationManager

def __main(args):
    if not os.path.exists(args.report_path):
        raise FunctionalException(f"Report path '{args.report_path}' doesn't exist")
    eh_path = os.path.join(args.report_path, "execution_historic.json")
    if not os.path.exists(eh_path):
        raise FunctionalException(f"Report path '{args.report_path}' doesn't contain an 'execution_historic.json' file")
    
    output_path = args.output_path
    if output_path is None:
        output_path = os.path.join(args.report_path, "scenario_duration_limits.csv")
        
    sdm = ScenarioDurationManager()
    sdm.import_execution_historic(eh_path)
    
    scenario_duration_limits = sdm.compute_scenario_duration_limits()
    sdm.create_file_scenario_duration_limits(output_path, scenario_duration_limits)
    
    logging.info(f"Finished")
    

if __name__ == "__main__":
    # Arguments
    descr = """Generate file with statistics on scenario duration.
    It needs as input the folder of the test report for which generating statistics.

    Example 1: Generate file with default duration limits
               
            python generate_scenario_duration_limits.py -r {path_to_report}
            
    Example 2: Generate file with default duration limits and specific output filename
    
            python generate_scenario_duration_limits.py -r {path_to_report} -o {path_to_report}/scenario_duration_limits.csv
            
        Note: With this "-o" parameter, it is equivalent to Example 1.
    """
    
            
    parser = argparse.ArgumentParser(description=descr, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-r', '--report', dest='report_path', required=True,
                   help='Path to the test report folder.')
    
    parser.add_argument('-o', '--output', dest='output_path', default=None,
                   help='Path to the output file. If not defined, a new file will be generated in report folder')

    args = parser.parse_args()

    logging.basicConfig(format = '{asctime:s} | {thread:-5d} | {levelname:5s} | {module:35s} | {message:s}', style = '{', level=logging.DEBUG, handlers=[logging.StreamHandler()])
    
    # Execute
    try:
        __main(args)
    except:
        logging.exception("An error occured")
        sys.exit(1)

    sys.exit(0)
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.actors.actions import Action
from holado_core.common.handlers.redo import Redo
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)



class Actor(object):
    """ Base class for actor.
    An actor defines which actions are possible and how to process them.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, module_name):
        self.__module_name = module_name
        self.__actor_by_module_name = {}
        
    def initialize(self):
        """
        Initialize actor
        """
        pass
    
    @property
    def module_name(self):
        return self.__module_name
    
    def _get_actor_for_module(self, name):
        """
        @param name Module name
        @return Module actor
        """
        if name not in self.__actor_by_module_name:
            self.__actor_by_module_name[name] = self._initialize_module(name)
        return self.__actor_by_module_name[name]
    
    def _initialize_module(self, name):
        """
        Initialize a new instance of given module name
        @param name Module name
        @return New module instance
        """
        raise TechnicalException(f"Unmanaged module '{name}'")
    
    def _call_action_from_module(self, finder_info, method_name, *args, **kwargs):
        if finder_info is not None and finder_info.module_name is not None:
            module = self._get_actor_for_module(finder_info.module_name)
            if hasattr(module, method_name):
                method = getattr(module, method_name)
            else:
                raise TechnicalException(f"Method '{method_name}' is missing in module '{finder_info.module_name}'")
            return method(*args, **kwargs)
        else:
            raise TechnicalException("No such action")

    def act_then(self, action_1, action_2):
        """
        @param action_1 First action
        @param action_2 Second action
        @return Action that process first action then second action
        """
        def func(*args, **kwargs):
            result_action_1 = action_1.execute(*args, **kwargs)
            return action_2.execute(result_action_1)
        return Action(f"{action_1.name} then {action_2.name}", func)
    
    def _act_with_redo(self, action):
        """
        Manage redo for given action
        @param action Action
        @return New action managing redo
        """
        raise TechnicalException(f"Unmanaged action type '{Typing.get_object_class_fullname(action)}'")
    
    def _get_redo(self, action, *args, **kwargs):
        """
        Redo an action
        @param action Action
        """
        class ActionRedo(Redo):
            def __init__(self):
                super().__init__(f"redo {action.name}")
                
            def _process(self):
                return action.execute(*args, **kwargs)
            
            def _get_waited_description(self):
                return action.name
            
        return ActionRedo()
    
    def redo_until(self, action, action_args, action_kwargs, value, timeout_seconds):
        """
        Redo until action result for given arguments is given value
        @param action Action
        @param action_args Action args
        @param action_kwargs Action kwargs
        @param value Value
        @param timeout_seconds Timeout in seconds (if None, default timeout is used)
        @return New action processing redo
        """
        def func():
            redo = self._get_redo(action, *action_args, **action_kwargs)
            redo.redo_until(value)
            if timeout_seconds is not None:
                redo.with_timeout(timeout_seconds)
            redo.execute()
            return None
            
        return Action(f"redo {action.name} until '{value}'", func)
    
    def redo_until_none(self, action, action_args, action_kwargs, timeout_seconds):
        """
        Redo until action result for given arguments is None
        @param action Action
        @param action_args Action args
        @param action_kwargs Action kwargs
        @param timeout_seconds Timeout in seconds (if None, default timeout is used)
        @return New action processing redo
        """
        def func():
            redo = self._get_redo(action, *action_args, **action_kwargs)
            redo.redo_until_none()
            if timeout_seconds is not None:
                redo.with_timeout(timeout_seconds)
            redo.execute()
            return None
            
        return Action(f"redo {action.name} until None", func)
    
    def redo_until_not_none(self, action, action_args, action_kwargs, timeout_seconds):
        """
        Redo until action result for given arguments is not None
        @param action Action
        @param action_args Action args
        @param action_kwargs Action kwargs
        @param timeout_seconds Timeout in seconds (if None, default timeout is used)
        @return New action processing redo
        """
        def func():
            redo = self._get_redo(action, *action_args, **action_kwargs)
            redo.redo_until_not_none()
            if timeout_seconds is not None:
                redo.with_timeout(timeout_seconds)
            redo.execute()
            return None
            
        return Action(f"redo {action.name} until not None", func)
    
    def redo_while(self, action, action_args, action_kwargs, value, timeout_seconds):
        """
        Redo while action result for given arguments is given value
        @param action Action
        @param action_args Action args
        @param action_kwargs Action kwargs
        @param value Value
        @param timeout_seconds Timeout in seconds (if None, default timeout is used)
        @return New action processing redo
        """
        def func():
            redo = self._get_redo(action, *action_args, **action_kwargs)
            redo.redo_while(value)
            if timeout_seconds is not None:
                redo.with_timeout(timeout_seconds)
            redo.execute()
            return None
            
        return Action(f"redo {action.name} while '{value}'", func)
    
    def redo_while_none(self, action, action_args, action_kwargs, timeout_seconds):
        """
        Redo while action result for given arguments is None
        @param action Action
        @param action_args Action args
        @param action_kwargs Action kwargs
        @param timeout_seconds Timeout in seconds (if None, default timeout is used)
        @return New action processing redo
        """
        def func():
            redo = self._get_redo(action, *action_args, **action_kwargs)
            redo.redo_while_none()
            if timeout_seconds is not None:
                redo.with_timeout(timeout_seconds)
            redo.execute()
            return None
            
        return Action(f"redo {action.name} while None", func)
    
    def redo_while_not_none(self, action, action_args, action_kwargs, timeout_seconds):
        """
        Redo while action result for given arguments is not None
        @param action Action
        @param action_args Action args
        @param action_kwargs Action kwargs
        @param timeout_seconds Timeout in seconds (if None, default timeout is used)
        @return New action processing redo
        """
        def func():
            redo = self._get_redo(action, *action_args, **action_kwargs)
            redo.redo_while_not_none()
            if timeout_seconds is not None:
                redo.with_timeout(timeout_seconds)
            redo.execute()
            return None
            
        return Action(f"redo {action.name} while not None", func)
    
    
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.finders.then_finder import ThenFinder
from holado_core.common.inspectors.element_inspector import ElementInspector
from holado_core.common.finders.after_in_tree_finder import AfterInTreeFinder
import abc
from holado_core.common.inspectors.tools.inspect_parameters import InspectParameters
from holado_core.common.inspectors.tools.inspect_context import InspectContext

logger = logging.getLogger(__name__)



class TreeInspector(ElementInspector):
    """ Base class for inspector on tree elements.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, module_name):
        super().__init__(module_name)
    
    """
    Find element with given finder, after element in tree found with given finder
    @param finder_first Finder for first element
    @param finder_after_first Finder for element to find after first one
    @param inspect_context Inspect context
    @param inspect_parameters Inspect parameters
    @return Finder
    """
    def get_finder_element_after(self, finder_first, finder_after_first, inspect_context:InspectContext=None, inspect_parameters:InspectParameters=None):
        description = f"{finder_after_first.element_description} after {finder_first.element_description} in tree"
        res = ThenFinder(description)
        
        # Find first element
        res.set_next_finder(finder_first)
        
        # Then find in tree
        res.set_next_finder( AfterInTreeFinder(finder_after_first, description) )
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_parent(self, inspect_context:InspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder for parent
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        raise NotImplementedError
    
    def get_finder_children(self, inspect_context:InspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder for children
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        raise NotImplementedError
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.tools.tools import Tools
from holado_yaml.yaml.yaml_client import YAMLClient


logger = logging.getLogger(__name__)

try:
    import ruamel.yaml  # @UnresolvedImport
    with_ruamel_yaml = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"YAML is not available with ruamel.yaml. Initialization failed on error: {exc}")
    with_ruamel_yaml = False



class RuamelYAMLClient(YAMLClient):
    """
    Client for actions on YAML files.
    """
    
    @classmethod
    def is_available(cls):
        return with_ruamel_yaml
    
    def __init__(self, name=None, client_type=None, **kwargs):
        super().__init__(name)
        
        if client_type is None:
            client_type = kwargs.get('typ')
        self.__internal_client = ruamel.yaml.YAML(typ=client_type, **kwargs)
    
    @property
    def internal_client(self):
        return self.__internal_client
    
    def load_io_file(self, file_like_object):
        res = self.internal_client.load(file_like_object)
        self._rm_style_info(res)
        return res
    
    def load_multiple_documents_io_file(self, file_like_object):
        res = list(self.internal_client.load_all(file_like_object))
        self._rm_style_info(res)
        return res
        
    def save_in_io_file(self, file_like_object, data, **kwargs):
        self.internal_client.dump(data, file_like_object, **kwargs)
    
    def _rm_style_info(self, d):
        """Remove style info so that fields order can be preserved if saved again.
        """
        if isinstance(d, dict):
            if hasattr(d, 'fa'):
                d.fa._flow_style = None
            for k, v in d.items():
                self._rm_style_info(k)
                self._rm_style_info(v)
        elif isinstance(d, list):
            if hasattr(d, 'fa'):
                d.fa._flow_style = None
            for elem in d:
                self._rm_style_info(elem)







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from builtins import object
import re
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class TextInspecter(object):

    def __init__(self):
        self.__variable_manager = None
        self.__expression_evaluator = None
        
        self.__registered_functions = {}
        
        self.__regex_function = re.compile(r"^(\w+)\((.*)\)$")
        self.__regex_to_be_interpreted_with_function = re.compile(r"^\$\{(\w+)\((.*)\)\}$")
        
    def initialize(self, var_manager, expression_evaluator):
        self.__variable_manager = var_manager
        self.__expression_evaluator = expression_evaluator
    
    @property
    def _expression_evaluator(self):
        return self.__expression_evaluator
    
    @property
    def _variable_manager(self):
        return self.__variable_manager
    
    def is_to_interpret(self, text):
        # r_match = self.__regex_to_be_interpreted.match(text)
        # if r_match:
        #     return True
        #
        # indexes = self.__search_interpret_section(text)
        # return indexes is not None
        return self.get_number_of_sections(text) > 0
    
    def get_number_of_sections(self, text):
        section_indexes = self._search_all_interpret_sections(text)
        return len(section_indexes)
    
    def _search_all_interpret_sections(self, str_to_interpret):
        m = None
        match_failed = False
        exc_msgs = []
        
        # Try to use regex to group "${" with "}" and "{" with "}", but without skipping "{{" and "}}"
        reg = self.__build_regex_matching_interpret_sections(str_to_interpret, do_match_semicolon_open=True, do_skip_double_semicolon=False)
        try:
            m = re.match(reg, str_to_interpret, re.DOTALL)
        except Exception as exc:
            exc_msgs.append(f"Error while matching string to built regex:\n    regex: [{reg}]\n   string: [{str_to_interpret}]\n    error: {exc}")
            match_failed = True
            
        # If failed, try to use regex to group "${" with "}" and "{" with "}", and skipping "{{" and "}}"
        if match_failed:
            match_failed = False
            reg = self.__build_regex_matching_interpret_sections(str_to_interpret, do_match_semicolon_open=True, do_skip_double_semicolon=True)
            try:
                m = re.match(reg, str_to_interpret, re.DOTALL)
            except Exception as exc:
                exc_msgs.append(f"Error while matching string to built regex:\n    regex: [{reg}]\n   string: [{str_to_interpret}]\n    error: {exc}")
                match_failed = True
            
        # If failed, try to use regex to group "${" with "}", but do not group "{" with "}"
        if match_failed:
            match_failed = False
            reg = self.__build_regex_matching_interpret_sections(str_to_interpret, do_match_semicolon_open=False, do_skip_double_semicolon=False)
            try:
                m = re.match(reg, str_to_interpret, re.DOTALL)
            except Exception as exc:
                exc_msgs.append(f"Error while matching string to built regex:\n    regex: [{reg}]\n   string: [{str_to_interpret}]\n    error: {exc}")
                match_failed = True
            
        if match_failed:
            msgs = Tools.indent_string(4, '\n'.join(exc_msgs))
            raise TechnicalException("Failed to search interpret sections for string [{str_to_interpret}]:\n{msgs}"\
                                     .format(str_to_interpret=str_to_interpret, msgs=msgs))
        elif m is None:
            raise TechnicalException(f"Failed to build a working regex:\n    regex: [{reg}]\n   string: [{str_to_interpret}]")
        
        # Find all "${"
        section_start_indexes = [m.start() for m in re.finditer(r'\$\{', str_to_interpret)]
        
        # Intersect both search
        return [r for r in m.regs[1:] if r[0] in section_start_indexes]
    
    def __build_regex_matching_interpret_sections(self, str_to_interpret, do_match_semicolon_open=False, do_skip_double_semicolon=False):
        res = str_to_interpret
        
        # First, replace byte sections
        m = re.search(r"b'[^']*'", res)
        while m:
            res = res[:m.start()] + '.'*(m.end()-m.start()) + res[m.end():]
            m = re.search(r"b'[^']*'", res)
        
        res = res.replace('(','.')\
                 .replace(')','.')\
                 .replace('${','(..')
        if do_skip_double_semicolon:
            res = res.replace('{{', '..')
            res = res.replace('}}','..')
        if do_match_semicolon_open:
            res = res.replace('{','(.')
        res = res.replace('}','.)')
        res = re.sub(r'[^()]', '.', res)
        return res
        
    # Note: this first implementation of _search_interpret_section works but is slower
    # def _search_interpret_section(self, text, start=0, excluded_indexes=None):
    #     if excluded_indexes is None:
    #         excluded_indexes = []
    #
    #     sub_text = text if start == 0 else text[start:]
    #     list_indexes = self._search_all_interpret_sections(sub_text)
    #     for indexes in list_indexes:
    #         res_indexes = indexes if start == 0 else tuple(v+start for v in indexes)
    #         if res_indexes not in excluded_indexes:
    #             return res_indexes
    #     return None
    def _search_interpret_section(self, text, start=0, excluded_indexes=None):
        if excluded_indexes is None:
            excluded_indexes = []
    
        offset = start
        ind_beg = None
        ind_end = None
    
        while True:
            # Find first section delimiters after offset
            if ind_beg is None:
                ind_beg = text.find("${", offset)
                if ind_beg < 0:
                    return None
            if ind_end is None:
                ind_end = text.find("}", ind_beg)
                if ind_end < 0:
                    return None
    
            # Find closest begin separator to end separator
            ind = text.find("${", ind_beg + 2)
            if ind > 0 and ind < ind_end:
                # Continue with next begin section
                ind_beg = ind
                continue
    
            # If section not in excluded ones, section is found
            if (ind_beg, ind_end+1) not in excluded_indexes:
                return (ind_beg, ind_end+1)
    
            # Try again with new offset
            offset = ind_end + 1
            ind_beg = None
            ind_end = None
    
        raise TechnicalException("Unexpected case")
        
    def _get_args(self, args_str):
        return args_str.split(',')
        
    def _has_function(self, text, with_interpreted=True):
        func_name, _ = self._get_function_and_args_strings(text, with_interpreted)
        if func_name is not None:
            func = self._get_function(func_name)
            return func is not None
        return False
        
    def _get_function_and_args_strings(self, text, with_interpreted=True):
        if with_interpreted:
            r_match = self.__regex_to_be_interpreted_with_function.match(text)
        else:
            r_match = self.__regex_function.match(text)
        if r_match:
            return r_match.group(1), r_match.group(2)
        else:
            return None, None
        
    def _get_function(self, func_name):
        if func_name in self.__registered_functions:
            return self.__registered_functions[func_name]
        else:
            return None
        
    def register_function(self, function_name, function):
        self.__registered_functions[function_name] = function
        
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.handlers.object import DeleteableObject
import xmltodict
from holado_system.system.filesystem.file import File
import re
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
from holado.common.context.session_context import SessionContext


def __get_path_manager():
    return SessionContext.instance().path_manager


class StreamXMLFile(DeleteableObject):
    """
    Manage a stream XML file.
    
    A stream XML file is a file with many root elements that are valid XML.
    It is useful for example to process workflows of data formatted in XML.
    An example in HolAdo is the test report DetailedScenarioReportBuilder with an XML file, 
    which adds in a single file a new scenario XML report after each scenario.
    """

    def __init__(self, path, auto_flush=True, do_open=False, **open_kwargs):
        super().__init__(f"stream XML file '{path}'")

        self.__file = File(path, auto_flush=auto_flush, do_open=do_open, **open_kwargs)
        self.__current_line = None
        self.__current_pos = None

    def __enter__(self):
        self.__file.__enter__()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__file.__exit__(exc_type, exc_val, exc_tb)
    
    @property
    def path(self):
        return self.__file.path

    @property
    def internal_file(self):
        return self.__file
    
    @property
    def is_open(self):
        return self.__file.is_open

    def open(self, **kwargs):
        self.__file.open(**kwargs)

    def close(self):
        self.__file.close()

    def read_element_as_dict(self, **parse_kwargs):
        # Extract lines of next element
        lines = []
        root_tag = None
        regex_tag_start = re.compile(r"<(?!/)([^ >]+)>")
        regex_tag_end = re.compile(r"</([^ >]+)>")
        level = 0
        is_found = False
        current_element_line = ""
        while not is_found:
            # Read next file line if needed
            if self.__current_line is None:
                line = self.__file.readline()
                if not line:
                    break
                self.__current_line = line
                self.__current_pos = 0
                current_element_line = ""
            
            # Find next tag
            m_start = regex_tag_start.search(self.__current_line, pos=self.__current_pos)
            m_end = regex_tag_end.search(self.__current_line, pos=self.__current_pos)
            #print(f"{root_tag=} ; {level=}  #  {m_start=} ; {m_end=}  #  {self.__current_pos=} in {self.__current_line=}")
            is_start = m_start is not None and (m_end is None or m_start.start() < m_end.start())
            is_end = m_end is not None and (m_start is None or m_end.start() < m_start.start())
            if is_start:
                current_end = m_start.end()
                if root_tag is None:
                    root_tag = m_start.group(1)
                else:
                    level += 1
            elif is_end:
                current_end = m_end.end()
                if level > 0:
                    level -= 1
                elif m_end.group(1) == root_tag:
                    root_tag = None
                    is_found = True
                else:
                    raise TechnicalException(f"Failed to read next element from stream XML file '{self.path}' at position {self.__file.tell()}: got root end element tag '{m_end.group(1)}' that doesn't match previous root start element tag '{root_tag}'")
            else:
                current_end = len(self.__current_line)
            
            # Append in current element line the text until tag
            current_element_line += self.__current_line[self.__current_pos:current_end]
            self.__current_pos = current_end
            
            # Add element line and reset current line
            if self.__current_pos >= len(self.__current_line) or is_found:
                lines.append(current_element_line)
                if self.__current_pos >= len(self.__current_line):
                    self.__current_line = None
        
        # Check error cases
        if root_tag is not None:
            raise TechnicalException(f"Failed to read next element from stream XML file '{self.path}': end of file is reached whereas a root end element tag '{root_tag}' is still expected")
        
        # Manage when end of file is reached without new element
        content = ''.join(lines).strip()
        if len(content) == 0:
            return None
        
        # Parse element as dict
        # print(f"Parsing:\n--------------------------------\n{content}\n--------------------------------")
        return xmltodict.parse(content, **parse_kwargs)
    
    def read_elements_as_dict_list(self, **parse_kwargs):
        res = []
        while True:
            element = self.read_element_as_dict(**parse_kwargs)
            if element is not None:
                res.append(element)
            else:
                break
        return res
    
    def write_element_dict(self, data, pretty=True, full_document=False, indent=Tools.indent_string(4, ''), **unparse_kwargs):
        unparse_kwargs['pretty'] = pretty
        unparse_kwargs['full_document'] = full_document
        unparse_kwargs['indent'] = indent
        
        content = xmltodict.unparse(data, **unparse_kwargs)
        self.__file.write(content)
        if pretty:
            self.__file.write('\n')
    
    def write_elements_dict_list(self, data_list, pretty=True, full_document=False, indent=Tools.indent_string(4, ''), **unparse_kwargs):
        for data in data_list:
            self.write_element_dict(data, pretty=pretty, full_document=full_document, indent=indent, **unparse_kwargs)


    @classmethod
    def read_stream_xml_file_content_as_dict_list(cls, path, open_kwargs=None, parse_kwargs=None):
        open_kwargs = open_kwargs if open_kwargs is not None else {}
        parse_kwargs = parse_kwargs if parse_kwargs is not None else {}
        
        file = StreamXMLFile(path, **open_kwargs)
        return file.read_elements_as_dict_list(**parse_kwargs)

    @classmethod
    def write_stream_xml_file_with_dict_list(cls, path, data_list, open_kwargs=None, unparse_kwargs=None):
        open_kwargs = open_kwargs if open_kwargs is not None else {}
        unparse_kwargs = unparse_kwargs if unparse_kwargs is not None else {}
        
        __get_path_manager().makedirs(path)
        
        file = StreamXMLFile(path, **open_kwargs)
        file.write_elements_dict_list(data_list, **unparse_kwargs)
        return file




